"""Module that returns appropriate sql query strings"""
from enum import Enum


class SubjectQueryColumns(Enum):
    """Expected columns in the subject query response."""

    ID = "id"
    CLASS_VALUES = "class_values"
    SEX = "sex"
    BIRTH_DATE = "birth_date"
    PATERNAL_ID = "paternal_id"
    PATERNAL_CLASS_VALUES = "paternal_class_values"
    MATERNAL_ID = "maternal_id"
    MATERNAL_CLASS_VALUES = "maternal_class_values"
    SPECIES_NAME = "species_name"
    GROUP_NAME = "group_name"
    GROUP_DESCRIPTION = "group_description"


class TaskSetQueryColumns(Enum):
    """Expected columns in the task set query response."""

    TASK_ID = "id"
    TASK_TYPE_ID = "task_type_id"
    DATE_START = "date_start"
    DATE_END = "date_end"
    INVESTIGATOR_ID = "investigator_id"
    TASK_OBJECT = "task_object"
    TYPE_NAME = "type_name"
    PROTOCOL_NUMBER = "protocol_number"


class LabTracksQueries:
    """Class to hold sql query strings for LabTracks"""

    @staticmethod
    def procedures_from_subject_id(subject_id: str) -> str:
        """
        Retrieves the information to populate metadata about subjects.

        Parameters
        ----------
        subject_id : str
            This is the id in the LabTracks Task_Set table

        Returns
        -------
        str
            SQL query that can be used to retrieve the data from LabTracks
            sqlserver db.
        """
        # TODO: parsers for comments/descriptions?
        return (
            "SELECT"
            f"    TS.ID AS {TaskSetQueryColumns.TASK_ID.value},"
            f"    TS.TASK_TYPE_ID AS {TaskSetQueryColumns.TASK_TYPE_ID.value},"
            f"    TT.TYPE_NAME AS {TaskSetQueryColumns.TYPE_NAME.value},"
            f"    TS.DATE_START AS {TaskSetQueryColumns.DATE_START.value},"
            f"    TS.DATE_END AS {TaskSetQueryColumns.DATE_END.value},"
            f"    TS.INVESTIGATOR_ID AS "
            f"    {TaskSetQueryColumns.INVESTIGATOR_ID.value},"
            f"    TSO.TASK_OBJECT AS {TaskSetQueryColumns.TASK_OBJECT.value},"
            f"    AP.PROTOCOL_NUMBER AS "
            f"    {TaskSetQueryColumns.PROTOCOL_NUMBER.value}"
            "  FROM TASK_SET TS"
            "    INNER JOIN TASK_SET_OBJECT TSO "
            "    ON TS.ID = TSO.TASK_ID"
            "    INNER JOIN ANIMALS_COMMON AC "
            "    ON TSO.TASK_OBJECT = AC.ID"
            "    INNER JOIN TASK_TYPE TT "
            "    ON TS.TASK_TYPE_ID = TT.ID"
            "    INNER JOIN ACUC_PROTOCOL AP "
            "    ON TS.ACUC_LINK_ID = AP.LINK_INDEX"
            f" WHERE AC.ID={subject_id};"
        )

    @staticmethod
    def subject_from_subject_id(subject_id: str) -> str:
        """
        Retrieves the information to populate metadata about subjects.

        Parameters
        ----------
        subject_id : str
            This is the id in the LabTracks ANIMALS_COMMON table

        Returns
        -------
            str
                SQL query that can be used to retrieve the data from LabTracks
                sqlserver db.

        """
        return (
            "SELECT"
            f"    AC.ID AS {SubjectQueryColumns.ID.value},"
            f"    AC.CLASS_VALUES AS {SubjectQueryColumns.CLASS_VALUES.value},"
            f"    AC.SEX AS {SubjectQueryColumns.SEX.value},"
            f"    AC.BIRTH_DATE AS {SubjectQueryColumns.BIRTH_DATE.value},"
            f"    PATERNAL.ID AS {SubjectQueryColumns.PATERNAL_ID.value},"
            f"    PATERNAL.CLASS_VALUES "
            f"      AS {SubjectQueryColumns.PATERNAL_CLASS_VALUES.value},"
            f"    MATERNAL.ID AS {SubjectQueryColumns.MATERNAL_ID.value},"
            f"    MATERNAL.CLASS_VALUES "
            f"      AS {SubjectQueryColumns.MATERNAL_CLASS_VALUES.value},"
            f"    S.SPECIES_NAME AS {SubjectQueryColumns.SPECIES_NAME.value},"
            f"    GROUPS.GROUP_NAME AS {SubjectQueryColumns.GROUP_NAME.value},"
            f"    G.GROUP_DESCRIPTION "
            f"      AS {SubjectQueryColumns.GROUP_DESCRIPTION.value}"
            "  FROM ANIMALS_COMMON AC"
            "    LEFT OUTER JOIN ANIMALS_COMMON MATERNAL"
            "    ON AC.MATERNAL_INDEX = MATERNAL.ID"
            "    LEFT OUTER JOIN ANIMALS_COMMON PATERNAL "
            "    ON AC.PATERNAL_INDEX = PATERNAL.ID"
            "    LEFT OUTER JOIN SPECIES S "
            "    ON AC.SPECIES_ID = S.ID"
            "    LEFT OUTER JOIN GROUPS AS G "
            "    ON AC.GROUP_ID = G.ID"
            "    LEFT OUTER JOIN GROUPS "
            "    ON MATERNAL.GROUP_ID = GROUPS.ID"
            f" WHERE AC.ID={subject_id};"
        )



"""Tests query builder methods."""

import unittest

from aind_metadata_service.labtracks.query_builder import (
    LabTracksQueries,
    SubjectQueryColumns,
    TaskSetQueryColumns,
)


class TestLabTracksQueryBuilder(unittest.TestCase):
    """Tests LabTracksQueries methods."""

    def test_subject_from_species(self):
        """Tests sql string is created correctly."""
        subject_id = "625464"
        actual_output = LabTracksQueries.subject_from_subject_id(subject_id)
        expected_output = (
            "SELECT"
            f"    AC.ID AS {SubjectQueryColumns.ID.value},"
            f"    AC.CLASS_VALUES AS {SubjectQueryColumns.CLASS_VALUES.value},"
            f"    AC.SEX AS {SubjectQueryColumns.SEX.value},"
            f"    AC.BIRTH_DATE AS {SubjectQueryColumns.BIRTH_DATE.value},"
            f"    PATERNAL.ID AS {SubjectQueryColumns.PATERNAL_ID.value},"
            f"    PATERNAL.CLASS_VALUES "
            f"      AS {SubjectQueryColumns.PATERNAL_CLASS_VALUES.value},"
            f"    MATERNAL.ID AS {SubjectQueryColumns.MATERNAL_ID.value},"
            f"    MATERNAL.CLASS_VALUES "
            f"      AS {SubjectQueryColumns.MATERNAL_CLASS_VALUES.value},"
            f"    S.SPECIES_NAME AS {SubjectQueryColumns.SPECIES_NAME.value},"
            f"    GROUPS.GROUP_NAME AS {SubjectQueryColumns.GROUP_NAME.value},"
            f"    G.GROUP_DESCRIPTION "
            f"      AS {SubjectQueryColumns.GROUP_DESCRIPTION.value}"
            "  FROM ANIMALS_COMMON AC"
            "    LEFT OUTER JOIN ANIMALS_COMMON MATERNAL"
            "    ON AC.MATERNAL_INDEX = MATERNAL.ID"
            "    LEFT OUTER JOIN ANIMALS_COMMON PATERNAL "
            "    ON AC.PATERNAL_INDEX = PATERNAL.ID"
            "    LEFT OUTER JOIN SPECIES S "
            "    ON AC.SPECIES_ID = S.ID"
            "    LEFT OUTER JOIN GROUPS AS G "
            "    ON AC.GROUP_ID = G.ID"
            "    LEFT OUTER JOIN GROUPS "
            "    ON MATERNAL.GROUP_ID = GROUPS.ID"
            f" WHERE AC.ID={subject_id};"
        )
        self.assertEqual(expected_output, actual_output)

    def test_procedures_from_id(self):
        """Tests sql string is created correctly."""
        subject_id = "625464"
        actual_output = LabTracksQueries.procedures_from_subject_id(subject_id)
        expected_output = (
            "SELECT"
            f"    TS.ID AS {TaskSetQueryColumns.TASK_ID.value},"
            f"    TS.TASK_TYPE_ID AS {TaskSetQueryColumns.TASK_TYPE_ID.value},"
            f"    TT.TYPE_NAME AS {TaskSetQueryColumns.TYPE_NAME.value},"
            f"    TS.DATE_START AS {TaskSetQueryColumns.DATE_START.value},"
            f"    TS.DATE_END AS {TaskSetQueryColumns.DATE_END.value},"
            f"    TS.INVESTIGATOR_ID AS "
            f"    {TaskSetQueryColumns.INVESTIGATOR_ID.value},"
            f"    TSO.TASK_OBJECT AS {TaskSetQueryColumns.TASK_OBJECT.value},"
            f"    AP.PROTOCOL_NUMBER AS "
            f"    {TaskSetQueryColumns.PROTOCOL_NUMBER.value}"
            "  FROM TASK_SET TS"
            "    INNER JOIN TASK_SET_OBJECT TSO "
            "    ON TS.ID = TSO.TASK_ID"
            "    INNER JOIN ANIMALS_COMMON AC "
            "    ON TSO.TASK_OBJECT = AC.ID"
            "    INNER JOIN TASK_TYPE TT "
            "    ON TS.TASK_TYPE_ID = TT.ID"
            "    INNER JOIN ACUC_PROTOCOL AP "
            "    ON TS.ACUC_LINK_ID = AP.LINK_INDEX"
            f" WHERE AC.ID={subject_id};"
        )
        self.assertEqual(expected_output, actual_output)


if __name__ == "__main__":
    unittest.main()



"""Module template was autogenerated from sharepoint list schema."""

from dataclasses import dataclass
from datetime import date, datetime
from enum import Enum
from typing import List, Optional

from aind_data_schema.procedures import (
    Anaesthetic,
    BrainInjection,
    CoordinateReferenceLocation,
    Craniotomy,
    CraniotomyType,
    FiberImplant,
    Headframe,
    InjectionMaterial,
    IontophoresisInjection,
    NanojectInjection,
    OphysProbe,
    ProbeName,
    Side,
    SubjectProcedure,
)
from aind_data_schema.subject import Sex

from aind_metadata_service.sharepoint.nsb2019.models import NSBList
from aind_metadata_service.sharepoint.nsb2019.models import (
    Procedure as NSBProcedure,
)


@dataclass
class HeadPostInfo:
    """Container for head post information"""

    headframe_type: Optional[str] = None
    headframe_part_number: Optional[str] = None
    well_type: Optional[str] = None
    well_part_number: Optional[str] = None


class InjectionType(Enum):
    """Enum class for Injection Types"""

    NANOJECT = "Nanoject"
    IONTOPHORESIS = "Iontophoresis"


class InjectionRound(Enum):
    """Enum class for Injection Rounds"""

    FIRST = "First"
    SECOND = "Second"


class MappedNSBList:
    """Mapped Fields in Sharepoint list"""

    def __init__(self, nsb: NSBList):
        """Class constructor"""
        self._nsb = nsb

    @staticmethod
    def _parse_basic_float_str(float_str: Optional[str]) -> Optional[float]:
        """Parse string representation of float such as '0.25'."""
        try:
            return None if float_str is None else float(float_str)
        except ValueError:
            return None

    @staticmethod
    def _parse_ap_str(_: Optional[str]) -> Optional[float]:
        """Parse AP String"""
        # TODO: Figure out how to parse AP fields
        return None

    @staticmethod
    def _parse_dv_str(_: Optional[str]) -> Optional[float]:
        """Parse dv String"""
        # TODO: Figure out how to parse DV fields
        return None

    @staticmethod
    def _parse_ml_str(_: Optional[str]) -> Optional[float]:
        """Parse ml string"""
        # TODO: Figure out how to parse ML fields
        return None

    @staticmethod
    def _parse_iso_dur_str(_: Optional[str]) -> Optional[float]:
        """Parse iso duration strings"""
        # TODO: Figure out how to parse Iso Duration fields
        return None

    def _parse_weight_str(self, weight_str: Optional[str]) -> Optional[float]:
        """Parse weight strings"""
        # Most entries are recorded as simple floats. There are some outliers.
        # But we can map those to None for the time being.
        return self._parse_basic_float_str(weight_str)

    @staticmethod
    def _parse_alt_time_str(_: Optional[str]) -> Optional[float]:
        """Parse alternating time strings"""
        # TODO: Figure out how to parse alternating time fields
        return None

    @staticmethod
    def _parse_angle_str(_: Optional[str]) -> Optional[float]:
        """Parse angle strings"""
        # TODO: Figure out how to parse angle fields
        return None

    @staticmethod
    def _parse_current_str(_: Optional[str]) -> Optional[float]:
        """Parse current strings"""
        # TODO: Figure out how to parse current fields
        return None

    @staticmethod
    def _parse_length_of_time_str(_: Optional[str]) -> Optional[float]:
        """Parse length of time strings"""
        # TODO: Figure out how to parse length of time fields
        return None

    @staticmethod
    def _parse_virus_strain_str(_: Optional[str]) -> Optional[str]:
        """Parse virus strain strings"""
        # TODO: Figure out how to parse virus strain fields
        return None

    @staticmethod
    def _parse_inj_vol_str(_: Optional[str]) -> Optional[float]:
        """Parse injection volume strings"""
        # TODO: Figure out how to parse injection volume fields
        return None

    @staticmethod
    def _parse_datetime_to_date(dt: Optional[datetime]) -> Optional[date]:
        """Parse date from datetime"""
        return None if dt is None else dt.date()

    @property
    def aind_age_at_injection(self) -> Optional[str]:
        """Maps age_at_injection to aind model"""
        return self._nsb.age_at_injection

    @property
    def aind_ap2nd_inj(self) -> Optional[float]:
        """Maps ap2nd_inj to aind model"""
        return self._parse_ap_str(self._nsb.ap2nd_inj)

    @property
    def aind_author_id(self) -> Optional[int]:
        """Maps author_id to aind model"""
        return self._nsb.author_id

    @property
    def aind_breg2_lamb(self) -> Optional[float]:
        """Maps breg2_lamb to aind model"""
        return self._parse_basic_float_str(self._nsb.breg2_lamb)

    @property
    def aind_cage(self) -> Optional[str]:
        """Maps cage to aind model"""
        return self._nsb.cage

    @property
    def aind_color_tag(self) -> Optional[str]:
        """Maps color_tag to aind model"""
        return self._nsb.color_tag

    @property
    def aind_compliance_asset_id(self) -> Optional[str]:
        """Maps compliance_asset_id to aind model"""
        return self._nsb.compliance_asset_id

    @property
    def aind_craniotomy_type(self) -> Optional[CraniotomyType]:
        """Maps craniotomy_type to aind model"""
        return (
            CraniotomyType.OTHER
            if self._nsb.craniotomy_type is None
            else {
                self._nsb.craniotomy_type.SELECT: CraniotomyType.OTHER,
                self._nsb.craniotomy_type.VISUAL_CORTEX_5MM: (
                    CraniotomyType.VISCTX
                ),
                self._nsb.craniotomy_type.FRONTAL_WINDOW_3MM: (
                    CraniotomyType.THREE_MM
                ),
                self._nsb.craniotomy_type.WHC_NP: CraniotomyType.WHC,
                self._nsb.craniotomy_type.WHC_2_P: CraniotomyType.WHC,
            }.get(self._nsb.craniotomy_type, CraniotomyType.OTHER)
        )

    @property
    def aind_created(self) -> Optional[datetime]:
        """Maps created to aind model"""
        return self._nsb.created

    @property
    def aind_date1st_injection(self) -> Optional[date]:
        """Maps date1st_injection to aind model"""
        return self._parse_datetime_to_date(self._nsb.date1st_injection)

    @property
    def aind_date2nd_injection(self) -> Optional[date]:
        """Maps date2nd_injection to aind model"""
        return self._parse_datetime_to_date(self._nsb.date2nd_injection)

    @property
    def aind_date_of_birth(self) -> Optional[date]:
        """Maps date_of_birth to aind model"""
        return self._parse_datetime_to_date(self._nsb.date_of_birth)

    @property
    def aind_date_of_surgery(self) -> Optional[date]:
        """Maps date_of_surgery to aind model"""
        return self._parse_datetime_to_date(self._nsb.date_of_surgery)

    @property
    def aind_date_range_end(self) -> Optional[datetime]:
        """Maps date_range_end to aind model"""
        return self._nsb.date_range_end

    @property
    def aind_date_range_start(self) -> Optional[datetime]:
        """Maps date_range_start to aind model"""
        return self._nsb.date_range_start

    @property
    def aind_dv2nd_inj(self) -> Optional[float]:
        """Maps dv2nd_inj to aind model"""
        return self._parse_dv_str(self._nsb.dv2nd_inj)

    @property
    def aind_editor_id(self) -> Optional[int]:
        """Maps editor_id to aind model"""
        return self._nsb.editor_id

    @property
    def aind_end_of_week(self) -> Optional[datetime]:
        """Maps end_of_week to aind model"""
        return self._nsb.end_of_week

    @property
    def aind_fiber_implant1(self) -> Optional[bool]:
        """Maps fiber_implant1 to aind model"""
        return self._nsb.fiber_implant1

    @property
    def aind_fiber_implant1_dv(self) -> Optional[float]:
        """Maps fiber_implant1_dv to aind model"""
        return self._parse_dv_str(self._nsb.fiber_implant1_dv)

    @property
    def aind_fiber_implant2(self) -> Optional[bool]:
        """Maps fiber_implant2 to aind model"""
        return self._nsb.fiber_implant2

    @property
    def aind_fiber_implant2_dv(self) -> Optional[float]:
        """Maps fiber_implant2_dv to aind model"""
        return self._parse_dv_str(self._nsb.fiber_implant2_dv)

    @property
    def aind_field30(self) -> Optional[str]:
        """Maps field30 to aind model"""
        return self._nsb.field30

    @property
    def aind_field50(self) -> Optional[str]:
        """Maps field50 to aind model"""
        return self._nsb.field50

    @property
    def aind_first_inj_recovery(self) -> Optional[float]:
        """Maps first_inj_recovery to aind model"""
        return self._nsb.first_inj_recovery

    @property
    def aind_first_injection_iso_durat(self) -> Optional[float]:
        """Maps first_injection_iso_durat to aind model"""
        return self._parse_iso_dur_str(self._nsb.first_injection_iso_durat)

    @property
    def aind_first_injection_weight_af(self) -> Optional[str]:
        """Maps first_injection_weight_af to aind model"""
        return self._parse_weight_str(self._nsb.first_injection_weight_af)

    @property
    def aind_first_injection_weight_be(self) -> Optional[str]:
        """Maps first_injection_weight_be to aind model"""
        return self._parse_weight_str(self._nsb.first_injection_weight_be)

    @property
    def aind_headpost_type(self) -> HeadPostInfo:
        """Maps headpost_type to aind model"""
        return (
            HeadPostInfo()
            if self._nsb.headpost_type is None
            else {
                self._nsb.headpost_type.SELECT: HeadPostInfo(),
                self._nsb.headpost_type.CAMSTYLE_HEADFRAME_016010: (
                    HeadPostInfo(
                        headframe_type="CAM-style",
                        headframe_part_number="0160-100-10 Rev A",
                        well_type="CAM-style",
                    )
                ),
                self._nsb.headpost_type.NEUROPIXELSTYLE_HEADFRAME: (
                    HeadPostInfo(
                        headframe_type="Neuropixel-style",
                        headframe_part_number="0160-100-10",
                        well_type="Neuropixel-style",
                        well_part_number="0160-200-36",
                    )
                ),
                self._nsb.headpost_type.MESOSCOPESTYLE_WELL_WITH: (
                    HeadPostInfo(
                        headframe_type="NGC-style",
                        headframe_part_number="0160-100-10",
                        well_type="Mesoscope-style",
                        well_part_number="0160-200-20",
                    )
                ),
                self._nsb.headpost_type.WHC_42_WITH_NEUROPIXEL_WE: (
                    HeadPostInfo(
                        headframe_type="WHC #42",
                        headframe_part_number="42",
                        well_type="Neuropixel-style",
                        well_part_number="0160-200-36",
                    )
                ),
                self._nsb.headpost_type.NGCSTYLE_HEADFRAME_NO_WEL: (
                    HeadPostInfo(
                        headframe_type="NGC-style",
                        headframe_part_number="0160-100-10",
                    )
                ),
                self._nsb.headpost_type.AI_STRAIGHT_HEADBAR: (
                    HeadPostInfo(headframe_type="AI Straight Headbar")
                ),
            }.get(self._nsb.headpost_type, HeadPostInfo())
        )

    @property
    def aind_hemisphere2nd_inj(self) -> Optional[Side]:
        """Maps hemisphere2nd_inj to aind model"""
        return (
            None
            if self._nsb.hemisphere2nd_inj is None
            else {
                self._nsb.hemisphere2nd_inj.SELECT: None,
                self._nsb.hemisphere2nd_inj.LEFT: Side.LEFT,
                self._nsb.hemisphere2nd_inj.RIGHT: Side.RIGHT,
            }.get(self._nsb.hemisphere2nd_inj, None)
        )

    @property
    def aind_hp_a_p(self) -> Optional[float]:
        """Maps hp_a_p to aind model"""
        return self._parse_ap_str(self._nsb.hp_a_p)

    @property
    def aind_hp_diameter(self) -> Optional[str]:
        """Maps hp_diameter to aind model"""
        return self._nsb.hp_diameter

    @property
    def aind_hp_durotomy(self) -> Optional[bool]:
        """Maps hp_durotomy to aind model"""
        return {
            self._nsb.hp_durotomy.SELECT: None,
            self._nsb.hp_durotomy.YES: True,
            self._nsb.hp_durotomy.NO: False,
        }.get(self._nsb.hp_durotomy, None)

    @property
    def aind_hp_inj(self) -> Optional[bool]:
        """Maps hp_inj to aind model"""
        return (
            True
            if self._nsb.hp_inj is not None and self._nsb.hp_inj == "Yes"
            else False
        )

    @property
    def aind_hp_iso_level(self) -> Optional[float]:
        """Maps hp_iso_level to aind model"""
        return (
            None
            if self._nsb.hp_iso_level is None
            else {
                self._nsb.hp_iso_level.SELECT: None,
                self._nsb.hp_iso_level.N_025: 0.25,
                self._nsb.hp_iso_level.N_050: 0.50,
                self._nsb.hp_iso_level.N_075: 0.75,
                self._nsb.hp_iso_level.N_100: 1.00,
                self._nsb.hp_iso_level.N_125: 1.25,
                self._nsb.hp_iso_level.N_15: 1.5,
                self._nsb.hp_iso_level.N_175: 1.75,
                self._nsb.hp_iso_level.N_200: 2.00,
                self._nsb.hp_iso_level.N_225: 2.25,
                self._nsb.hp_iso_level.N_250: 2.50,
                self._nsb.hp_iso_level.N_275: 2.75,
                self._nsb.hp_iso_level.N_300: None,
            }.get(self._nsb.hp_iso_level, None)
        )

    @property
    def aind_hp_loc(self) -> Optional[Side]:
        """Maps hp_loc to aind model"""
        return (
            None
            if self._nsb.hp_loc is None
            else {
                self._nsb.hp_loc.SELECT: None,
                self._nsb.hp_loc.LEFT: Side.LEFT,
                self._nsb.hp_loc.CENTER: None,
                self._nsb.hp_loc.RIGHT: Side.RIGHT,
            }.get(self._nsb.hp_loc, None)
        )

    @property
    def aind_hp_m_l(self) -> Optional[float]:
        """Maps hp_m_l to aind model"""
        return self._parse_ml_str(self._nsb.hp_m_l)

    @property
    def aind_hp_recovery(self) -> Optional[float]:
        """Maps hp_recovery to aind model"""
        return self._nsb.hp_recovery

    @property
    def aind_hp_requestor(self) -> Optional[str]:
        """Maps hp_requestor to aind model"""
        return self._nsb.hp_requestor

    @property
    def aind_hp_requestor_comments_pla(self) -> Optional[str]:
        """Maps hp_requestor_comments_pla to aind model"""
        return self._nsb.hp_requestor_comments_pla

    @property
    def aind_hp_surgeon_comments(self) -> Optional[str]:
        """Maps hp_surgeon_comments to aind model"""
        return self._nsb.hp_surgeon_comments

    @property
    def aind_hp_work_station(self) -> Optional[str]:
        """Maps hp_work_station to aind model"""
        return (
            None
            if self._nsb.hp_work_station is None
            else {
                self._nsb.hp_work_station.SELECT: None,
                self._nsb.hp_work_station.SWS_1: (
                    self._nsb.hp_work_station.SWS_1.value
                ),
                self._nsb.hp_work_station.SWS_2: (
                    self._nsb.hp_work_station.SWS_2.value
                ),
                self._nsb.hp_work_station.SWS_3: (
                    self._nsb.hp_work_station.SWS_3.value
                ),
                self._nsb.hp_work_station.SWS_4: (
                    self._nsb.hp_work_station.SWS_4.value
                ),
                self._nsb.hp_work_station.SWS_5: (
                    self._nsb.hp_work_station.SWS_5.value
                ),
                self._nsb.hp_work_station.SWS_6: (
                    self._nsb.hp_work_station.SWS_6.value
                ),
                self._nsb.hp_work_station.SWS_7: (
                    self._nsb.hp_work_station.SWS_7.value
                ),
                self._nsb.hp_work_station.SWS_8: (
                    self._nsb.hp_work_station.SWS_8.value
                ),
                self._nsb.hp_work_station.SWS_9: (
                    self._nsb.hp_work_station.SWS_9.value
                ),
            }.get(self._nsb.hp_work_station, None)
        )

    @property
    def aind_iacuc_protocol(self) -> Optional[str]:
        """Maps iacuc_protocol to aind model"""
        return (
            None
            if self._nsb.iacuc_protocol is None
            else {
                self._nsb.iacuc_protocol.SELECT: None,
                self._nsb.iacuc_protocol.N_2001: (
                    self._nsb.iacuc_protocol.N_2001.value
                ),
                self._nsb.iacuc_protocol.N_2002: (
                    self._nsb.iacuc_protocol.N_2002.value
                ),
                self._nsb.iacuc_protocol.N_2003: (
                    self._nsb.iacuc_protocol.N_2003.value
                ),
                self._nsb.iacuc_protocol.N_2004: (
                    self._nsb.iacuc_protocol.N_2004.value
                ),
                self._nsb.iacuc_protocol.N_2005: (
                    self._nsb.iacuc_protocol.N_2005.value
                ),
                self._nsb.iacuc_protocol.N_2006: (
                    self._nsb.iacuc_protocol.N_2006.value
                ),
                self._nsb.iacuc_protocol.N_2011: (
                    self._nsb.iacuc_protocol.N_2011.value
                ),
                self._nsb.iacuc_protocol.N_2102: (
                    self._nsb.iacuc_protocol.N_2102.value
                ),
                self._nsb.iacuc_protocol.N_2103: (
                    self._nsb.iacuc_protocol.N_2103.value
                ),
                self._nsb.iacuc_protocol.N_2104: (
                    self._nsb.iacuc_protocol.N_2104.value
                ),
                self._nsb.iacuc_protocol.N_2105: (
                    self._nsb.iacuc_protocol.N_2105.value
                ),
                self._nsb.iacuc_protocol.N_2106: (
                    self._nsb.iacuc_protocol.N_2106.value
                ),
                self._nsb.iacuc_protocol.N_2107: (
                    self._nsb.iacuc_protocol.N_2107.value
                ),
                self._nsb.iacuc_protocol.N_2108: (
                    self._nsb.iacuc_protocol.N_2108.value
                ),
                self._nsb.iacuc_protocol.N_2109: (
                    self._nsb.iacuc_protocol.N_2109.value
                ),
                self._nsb.iacuc_protocol.N_2110: (
                    self._nsb.iacuc_protocol.N_2110.value
                ),
                self._nsb.iacuc_protocol.N_2113: (
                    self._nsb.iacuc_protocol.N_2113.value
                ),
                self._nsb.iacuc_protocol.N_2115: (
                    self._nsb.iacuc_protocol.N_2115.value
                ),
                self._nsb.iacuc_protocol.N_2117: (
                    self._nsb.iacuc_protocol.N_2117.value
                ),
                self._nsb.iacuc_protocol.N_2201: (
                    self._nsb.iacuc_protocol.N_2201.value
                ),
                self._nsb.iacuc_protocol.N_2202: (
                    self._nsb.iacuc_protocol.N_2202.value
                ),
                self._nsb.iacuc_protocol.N_2205: (
                    self._nsb.iacuc_protocol.N_2205.value
                ),
                self._nsb.iacuc_protocol.N_2212: (
                    self._nsb.iacuc_protocol.N_2212.value
                ),
            }.get(self._nsb.iacuc_protocol, None)
        )

    @property
    def aind_id(self) -> Optional[int]:
        """Maps id to aind model"""
        return self._nsb.id

    @property
    def aind_inj1_alternating_time(self) -> Optional[float]:
        """Maps inj1_alternating_time to aind model"""
        return self._parse_alt_time_str(self._nsb.inj1_alternating_time)

    @property
    def aind_inj1_angle_v2(self) -> Optional[float]:
        """Maps inj1_angle_v2 to aind model"""
        return self._parse_angle_str(self._nsb.inj1_angle_v2)

    @property
    def aind_inj1_current(self) -> Optional[float]:
        """Maps inj1_current to aind model"""
        return self._parse_current_str(self._nsb.inj1_current)

    @property
    def aind_inj1_lenghtof_time(self) -> Optional[float]:
        """Maps inj1_lenghtof_time to aind model"""
        return self._parse_length_of_time_str(self._nsb.inj1_lenghtof_time)

    @property
    def aind_inj1_round(self) -> Optional[InjectionRound]:
        """Maps inj1_round to aind model"""
        return (
            None
            if self._nsb.inj1_round is None
            else {
                self._nsb.inj1_round.SELECT: None,
                self._nsb.inj1_round.N_1ST: InjectionRound.FIRST,
                self._nsb.inj1_round.N_2ND: InjectionRound.SECOND,
                self._nsb.inj1_round.NA: None,
            }.get(self._nsb.inj1_round, None)
        )

    @property
    def aind_inj1_storage_location(self) -> Optional[str]:
        """Maps inj1_storage_location to aind model"""
        return self._nsb.inj1_storage_location

    @property
    def aind_inj1_type(self) -> Optional[InjectionType]:
        """Maps inj1_type to aind model"""
        return (
            None
            if self._nsb.inj1_type is None
            else {
                self._nsb.inj1_type.SELECT: None,
                self._nsb.inj1_type.IONTOPHORESIS: InjectionType.IONTOPHORESIS,
                self._nsb.inj1_type.NANOJECT_PRESSURE: InjectionType.NANOJECT,
            }.get(self._nsb.inj1_type, None)
        )

    @property
    def aind_inj1_virus_strain_rt(self) -> Optional[str]:
        """Maps inj1_virus_strain_rt to aind model"""
        return self._parse_virus_strain_str(self._nsb.inj1_virus_strain_rt)

    @property
    def aind_inj1_vol(self) -> Optional[str]:
        """Maps inj1_vol to aind model"""
        return self._parse_inj_vol_str(self._nsb.inj1_vol)

    @property
    def aind_inj1angle0(self) -> Optional[float]:
        """Maps inj1angle0 to aind model"""
        return (
            None
            if self._nsb.inj1angle0 is None
            else {
                self._nsb.inj1angle0.SELECT: None,
                self._nsb.inj1angle0.N_0_DEGREES: 0,
                self._nsb.inj1angle0.N_10_DEGREES: 10,
                self._nsb.inj1angle0.N_15_DEGREES: 15,
                self._nsb.inj1angle0.N_20_DEGREES: 20,
                self._nsb.inj1angle0.N_30_DEGREES: 30,
                self._nsb.inj1angle0.N_40_DEGREES: 40,
            }.get(self._nsb.inj1angle0, None)
        )

    @property
    def aind_inj1volperdepth(self) -> Optional[float]:
        """Maps inj1volperdepth to aind model"""
        return self._parse_inj_vol_str(self._nsb.inj1volperdepth)

    @property
    def aind_inj2_alternating_time(self) -> Optional[float]:
        """Maps inj2_alternating_time to aind model"""
        return self._parse_alt_time_str(self._nsb.inj2_alternating_time)

    @property
    def aind_inj2_angle_v2(self) -> Optional[float]:
        """Maps inj2_angle_v2 to aind model"""
        return self._parse_angle_str(self._nsb.inj2_angle_v2)

    @property
    def aind_inj2_current(self) -> Optional[float]:
        """Maps inj2_current to aind model"""
        return self._parse_current_str(self._nsb.inj2_current)

    @property
    def aind_inj2_lenghtof_time(self) -> Optional[float]:
        """Maps inj2_lenghtof_time to aind model"""
        return self._parse_length_of_time_str(self._nsb.inj2_lenghtof_time)

    @property
    def aind_inj2_round(self) -> Optional[InjectionRound]:
        """Maps inj2_round to aind model"""
        return (
            None
            if self._nsb.inj2_round is None
            else {
                self._nsb.inj2_round.SELECT: None,
                self._nsb.inj2_round.N_1ST: InjectionRound.FIRST,
                self._nsb.inj2_round.N_2ND: InjectionRound.SECOND,
                self._nsb.inj2_round.NA: None,
            }.get(self._nsb.inj2_round, None)
        )

    @property
    def aind_inj2_storage_location(self) -> Optional[str]:
        """Maps inj2_storage_location to aind model"""
        return self._nsb.inj2_storage_location

    @property
    def aind_inj2_type(self) -> Optional[InjectionType]:
        """Maps inj2_type to aind model"""
        return (
            None
            if self._nsb.inj2_type is None
            else {
                self._nsb.inj2_type.SELECT: None,
                self._nsb.inj2_type.IONTOPHORESIS: InjectionType.IONTOPHORESIS,
                self._nsb.inj2_type.NANOJECT_PRESSURE: InjectionType.NANOJECT,
            }.get(self._nsb.inj2_type, None)
        )

    @property
    def aind_inj2_virus_strain_rt(self) -> Optional[str]:
        """Maps inj2_virus_strain_rt to aind model"""
        return self._parse_virus_strain_str(self._nsb.inj2_virus_strain_rt)

    @property
    def aind_inj2_vol(self) -> Optional[float]:
        """Maps inj2_vol to aind model"""
        return self._parse_inj_vol_str(self._nsb.inj2_vol)

    @property
    def aind_inj2angle0(self) -> Optional[float]:
        """Maps inj2angle0 to aind model"""
        return (
            None
            if self._nsb.inj2angle0 is None
            else {
                self._nsb.inj2angle0.SELECT: None,
                self._nsb.inj2angle0.N_0_DEGREES: 0,
                self._nsb.inj2angle0.N_10_DEGREES: 10,
                self._nsb.inj2angle0.N_15_DEGREES: 15,
                self._nsb.inj2angle0.N_20_DEGREES: 20,
                self._nsb.inj2angle0.N_30_DEGREES: 30,
                self._nsb.inj2angle0.N_40_DEGREES: 40,
            }.get(self._nsb.inj2angle0, None)
        )

    @property
    def aind_inj2volperdepth(self) -> Optional[float]:
        """Maps inj2volperdepth to aind model"""
        return self._parse_inj_vol_str(self._nsb.inj2volperdepth)

    @property
    def aind_ionto_number_hpinj(self) -> Optional[str]:
        """Maps ionto_number_hpinj to aind model"""
        return (
            None
            if self._nsb.ionto_number_hpinj is None
            else {
                self._nsb.ionto_number_hpinj.IONTO_1: (
                    self._nsb.ionto_number_hpinj.IONTO_1.value
                ),
                self._nsb.ionto_number_hpinj.IONTO_2: (
                    self._nsb.ionto_number_hpinj.IONTO_2.value
                ),
                self._nsb.ionto_number_hpinj.IONTO_3: (
                    self._nsb.ionto_number_hpinj.IONTO_3.value
                ),
                self._nsb.ionto_number_hpinj.IONTO_4: (
                    self._nsb.ionto_number_hpinj.IONTO_4.value
                ),
                self._nsb.ionto_number_hpinj.IONTO_5: (
                    self._nsb.ionto_number_hpinj.IONTO_5.value
                ),
                self._nsb.ionto_number_hpinj.IONTO_6: (
                    self._nsb.ionto_number_hpinj.IONTO_6.value
                ),
                self._nsb.ionto_number_hpinj.IONTO_7: (
                    self._nsb.ionto_number_hpinj.IONTO_7.value
                ),
                self._nsb.ionto_number_hpinj.IONTO_8: (
                    self._nsb.ionto_number_hpinj.IONTO_8.value
                ),
                self._nsb.ionto_number_hpinj.IONTO_9: (
                    self._nsb.ionto_number_hpinj.IONTO_9.value
                ),
            }.get(self._nsb.ionto_number_hpinj, None)
        )

    @property
    def aind_ionto_number_inj1(self) -> Optional[str]:
        """Maps ionto_number_inj1 to aind model"""
        return (
            None
            if self._nsb.ionto_number_inj1 is None
            else {
                self._nsb.ionto_number_inj1.SELECT: None,
                self._nsb.ionto_number_inj1.IONTO_1: (
                    self._nsb.ionto_number_inj1.IONTO_1.value
                ),
                self._nsb.ionto_number_inj1.IONTO_2: (
                    self._nsb.ionto_number_inj1.IONTO_2.value
                ),
                self._nsb.ionto_number_inj1.IONTO_3: (
                    self._nsb.ionto_number_inj1.IONTO_3.value
                ),
                self._nsb.ionto_number_inj1.IONTO_4: (
                    self._nsb.ionto_number_inj1.IONTO_4.value
                ),
                self._nsb.ionto_number_inj1.IONTO_5: (
                    self._nsb.ionto_number_inj1.IONTO_5.value
                ),
                self._nsb.ionto_number_inj1.IONTO_6: (
                    self._nsb.ionto_number_inj1.IONTO_6.value
                ),
                self._nsb.ionto_number_inj1.IONTO_7: (
                    self._nsb.ionto_number_inj1.IONTO_7.value
                ),
                self._nsb.ionto_number_inj1.IONTO_8: (
                    self._nsb.ionto_number_inj1.IONTO_8.value
                ),
                self._nsb.ionto_number_inj1.IONTO_9: (
                    self._nsb.ionto_number_inj1.IONTO_9.value
                ),
                self._nsb.ionto_number_inj1.IONTO_10: (
                    self._nsb.ionto_number_inj1.IONTO_10.value
                ),
                self._nsb.ionto_number_inj1.NA: None,
            }.get(self._nsb.ionto_number_inj1, None)
        )

    @property
    def aind_ionto_number_inj2(self) -> Optional[str]:
        """Maps ionto_number_inj2 to aind model"""
        return (
            None
            if self._nsb.ionto_number_inj2 is None
            else {
                self._nsb.ionto_number_inj2.SELECT: None,
                self._nsb.ionto_number_inj2.IONTO_1: (
                    self._nsb.ionto_number_inj2.IONTO_1.value
                ),
                self._nsb.ionto_number_inj2.IONTO_2: (
                    self._nsb.ionto_number_inj2.IONTO_2.value
                ),
                self._nsb.ionto_number_inj2.IONTO_3: (
                    self._nsb.ionto_number_inj2.IONTO_3.value
                ),
                self._nsb.ionto_number_inj2.IONTO_4: (
                    self._nsb.ionto_number_inj2.IONTO_4.value
                ),
                self._nsb.ionto_number_inj2.IONTO_5: (
                    self._nsb.ionto_number_inj2.IONTO_5.value
                ),
                self._nsb.ionto_number_inj2.IONTO_6: (
                    self._nsb.ionto_number_inj2.IONTO_6.value
                ),
                self._nsb.ionto_number_inj2.IONTO_7: (
                    self._nsb.ionto_number_inj2.IONTO_7.value
                ),
                self._nsb.ionto_number_inj2.IONTO_8: (
                    self._nsb.ionto_number_inj2.IONTO_8.value
                ),
                self._nsb.ionto_number_inj2.IONTO_9: (
                    self._nsb.ionto_number_inj2.IONTO_9.value
                ),
                self._nsb.ionto_number_inj2.IONTO_10: (
                    self._nsb.ionto_number_inj2.IONTO_10.value
                ),
                self._nsb.ionto_number_inj2.NA: None,
            }.get(self._nsb.ionto_number_inj2, None)
        )

    @property
    def aind_iso_on(self) -> Optional[float]:
        """Maps iso_on to aind model"""
        return self._nsb.iso_on

    @property
    def aind_lab_tracks_group(self) -> Optional[str]:
        """Maps lab_tracks_group to aind model"""
        return self._nsb.lab_tracks_group

    @property
    def aind_lab_tracks_id(self) -> Optional[str]:
        """Maps lab_tracks_id to aind model"""
        return self._nsb.lab_tracks_id

    @property
    def aind_lab_tracks_requestor(self) -> Optional[str]:
        """Maps lab_tracks_requestor to aind model"""
        return self._nsb.lab_tracks_requestor

    @property
    def aind_lims_link(self) -> Optional[str]:
        """Maps lims_link to aind model"""
        return self._nsb.lims_link

    @property
    def aind_long1st_round_inj_cmts(self) -> Optional[str]:
        """Maps long1st_round_inj_cmts to aind model"""
        return self._nsb.long1st_round_inj_cmts

    @property
    def aind_long2nd_rnd_inj_cmts(self) -> Optional[str]:
        """Maps long2nd_rnd_inj_cmts to aind model"""
        return self._nsb.long2nd_rnd_inj_cmts

    @property
    def aind_long_requestor_comments(self) -> Optional[str]:
        """Maps long_requestor_comments to aind model"""
        return self._nsb.long_requestor_comments

    @property
    def aind_long_surgeon_comments(self) -> Optional[str]:
        """Maps long_surgeon_comments to aind model"""
        return self._nsb.long_surgeon_comments

    @property
    def aind_ml2nd_inj(self) -> Optional[float]:
        """Maps ml2nd_inj to aind model"""
        return self._parse_ml_str(self._nsb.ml2nd_inj)

    @property
    def aind_modified(self) -> Optional[datetime]:
        """Maps modified to aind model"""
        return self._nsb.modified

    @property
    def aind_nanoject_number_inj10(self) -> Optional[str]:
        """Maps nanoject_number_inj10 to aind model"""
        return (
            None
            if self._nsb.nanoject_number_inj10 is None
            else {
                self._nsb.nanoject_number_inj10.SELECT: None,
                self._nsb.nanoject_number_inj10.NJ1: (
                    self._nsb.nanoject_number_inj10.NJ1.value
                ),
                self._nsb.nanoject_number_inj10.NJ2: (
                    self._nsb.nanoject_number_inj10.NJ2.value
                ),
                self._nsb.nanoject_number_inj10.NJ3: (
                    self._nsb.nanoject_number_inj10.NJ3.value
                ),
                self._nsb.nanoject_number_inj10.NJ4: (
                    self._nsb.nanoject_number_inj10.NJ4.value
                ),
                self._nsb.nanoject_number_inj10.NJ5: (
                    self._nsb.nanoject_number_inj10.NJ5.value
                ),
                self._nsb.nanoject_number_inj10.NJ6: (
                    self._nsb.nanoject_number_inj10.NJ6.value
                ),
                self._nsb.nanoject_number_inj10.NJ7: (
                    self._nsb.nanoject_number_inj10.NJ7.value
                ),
                self._nsb.nanoject_number_inj10.NJ8: (
                    self._nsb.nanoject_number_inj10.NJ8.value
                ),
                self._nsb.nanoject_number_inj10.NA: None,
            }.get(self._nsb.nanoject_number_inj10, None)
        )

    @property
    def aind_nanoject_number_inj2(self) -> Optional[str]:
        """Maps nanoject_number_inj2 to aind model"""
        return (
            None
            if self._nsb.nanoject_number_inj2 is None
            else {
                self._nsb.nanoject_number_inj2.SELECT: None,
                self._nsb.nanoject_number_inj2.NJ1: (
                    self._nsb.nanoject_number_inj2.NJ1
                ),
                self._nsb.nanoject_number_inj2.NJ2: (
                    self._nsb.nanoject_number_inj2.NJ2
                ),
                self._nsb.nanoject_number_inj2.NJ3: (
                    self._nsb.nanoject_number_inj2.NJ3
                ),
                self._nsb.nanoject_number_inj2.NJ4: (
                    self._nsb.nanoject_number_inj2.NJ4
                ),
                self._nsb.nanoject_number_inj2.NJ5: (
                    self._nsb.nanoject_number_inj2.NJ5
                ),
                self._nsb.nanoject_number_inj2.NJ6: (
                    self._nsb.nanoject_number_inj2.NJ6
                ),
                self._nsb.nanoject_number_inj2.NJ7: (
                    self._nsb.nanoject_number_inj2.NJ7
                ),
                self._nsb.nanoject_number_inj2.NJ8: (
                    self._nsb.nanoject_number_inj2.NJ8
                ),
                self._nsb.nanoject_number_inj2.NA: None,
            }.get(self._nsb.nanoject_number_inj2, None)
        )

    @property
    def aind_nd_roung_injection_commen(self) -> Optional[str]:
        """Maps nd_roung_injection_commen to aind model"""
        return self._nsb.nd_roung_injection_commen

    @property
    def aind_pedigree_name(self) -> Optional[str]:
        """Maps pedigree_name to aind model"""
        return self._nsb.pedigree_name

    @property
    def aind_procedure(self) -> Optional[NSBProcedure]:
        """Maps procedure to aind model"""
        return self._nsb.procedure

    @property
    def aind_round1_inj_isolevel(self) -> Optional[float]:
        """Maps round1_inj_isolevel to aind model"""
        return (
            None
            if self._nsb.round1_inj_isolevel is None
            else {
                self._nsb.round1_inj_isolevel.SELECT: None,
                self._nsb.round1_inj_isolevel.N_025: 0.25,
                self._nsb.round1_inj_isolevel.N_050: 0.50,
                self._nsb.round1_inj_isolevel.N_075: 0.75,
                self._nsb.round1_inj_isolevel.N_100: 1.00,
                self._nsb.round1_inj_isolevel.N_125: 1.25,
                self._nsb.round1_inj_isolevel.N_150: 1.50,
                self._nsb.round1_inj_isolevel.N_175: 1.75,
                self._nsb.round1_inj_isolevel.N_200: 2.00,
                self._nsb.round1_inj_isolevel.N_225: 2.25,
                self._nsb.round1_inj_isolevel.N_250: 2.50,
                self._nsb.round1_inj_isolevel.N_275: 2.75,
                self._nsb.round1_inj_isolevel.N_300: None,
            }.get(self._nsb.round1_inj_isolevel, None)
        )

    @property
    def aind_round2_inj_isolevel(self) -> Optional[float]:
        """Maps round2_inj_isolevel to aind model"""
        return (
            None
            if self._nsb.round2_inj_isolevel is None
            else {
                self._nsb.round2_inj_isolevel.SELECT: None,
                self._nsb.round2_inj_isolevel.N_025: 0.25,
                self._nsb.round2_inj_isolevel.N_050: 0.50,
                self._nsb.round2_inj_isolevel.N_075: 0.75,
                self._nsb.round2_inj_isolevel.N_100: 1.00,
                self._nsb.round2_inj_isolevel.N_125: 1.25,
                self._nsb.round2_inj_isolevel.N_150: 1.50,
                self._nsb.round2_inj_isolevel.N_175: 1.75,
                self._nsb.round2_inj_isolevel.N_200: 2.00,
                self._nsb.round2_inj_isolevel.N_225: 2.25,
                self._nsb.round2_inj_isolevel.N_250: 2.50,
                self._nsb.round2_inj_isolevel.N_275: 2.75,
                self._nsb.round2_inj_isolevel.N_300: None,
            }.get(self._nsb.round2_inj_isolevel, None)
        )

    @property
    def aind_second_inj_recover(self) -> Optional[float]:
        """Maps second_inj_recover to aind model"""
        return self._nsb.second_inj_recover

    @property
    def aind_second_injection_iso_dura(self) -> Optional[float]:
        """Maps second_injection_iso_dura to aind model"""
        return self._parse_iso_dur_str(self._nsb.second_injection_iso_dura)

    @property
    def aind_second_injection_weight_a(self) -> Optional[float]:
        """Maps second_injection_weight_a to aind model"""
        return self._parse_weight_str(self._nsb.second_injection_weight_a)

    @property
    def aind_second_injection_weight_b(self) -> Optional[float]:
        """Maps second_injection_weight_b to aind model"""
        return self._parse_weight_str(self._nsb.second_injection_weight_b)

    @property
    def aind_sex(self) -> Optional[Sex]:
        """Maps sex to aind model"""
        return (
            None
            if self._nsb.sex is None
            else {
                self._nsb.sex.SELECT: None,
                self._nsb.sex.MALE: Sex.MALE,
                self._nsb.sex.FEMALE: Sex.FEMALE,
            }.get(self._nsb.sex, None)
        )

    @property
    def aind_st_round_injection_commen(self) -> Optional[str]:
        """Maps st_round_injection_commen to aind model"""
        return self._nsb.st_round_injection_commen

    @property
    def aind_start_of_week(self) -> Optional[datetime]:
        """Maps start_of_week to aind model"""
        return self._nsb.start_of_week

    @property
    def aind_title(self) -> Optional[str]:
        """Maps title to aind model"""
        return self._nsb.title

    @property
    def aind_touch_up_comp(self) -> Optional[datetime]:
        """Maps touch_up_comp to aind model"""
        return self._nsb.touch_up_comp

    @property
    def aind_touch_up_weight(self) -> Optional[float]:
        """Maps touch_up_weight to aind model"""
        return self._parse_weight_str(self._nsb.touch_up_weight)

    @property
    def aind_ui_version_string(self) -> Optional[str]:
        """Maps ui_version_string to aind model"""
        return self._nsb.ui_version_string

    @property
    def aind_virus_a_p(self) -> Optional[float]:
        """Maps virus_a_p to aind model"""
        return self._parse_ap_str(self._nsb.virus_a_p)

    @property
    def aind_virus_d_v(self) -> Optional[float]:
        """Maps virus_d_v to aind model"""
        return self._parse_dv_str(self._nsb.virus_d_v)

    @property
    def aind_virus_hemisphere(self) -> Optional[Side]:
        """Maps virus_hemisphere to aind model"""
        return (
            None
            if self._nsb.virus_hemisphere is None
            else {
                self._nsb.virus_hemisphere.SELECT: None,
                self._nsb.virus_hemisphere.LEFT: Side.LEFT,
                self._nsb.virus_hemisphere.RIGHT: Side.RIGHT,
            }.get(self._nsb.virus_hemisphere, None)
        )

    @property
    def aind_virus_m_l(self) -> Optional[float]:
        """Maps virus_m_l to aind model"""
        return self._parse_ml_str(self._nsb.virus_m_l)

    @property
    def aind_weight_after_surgery(self) -> Optional[float]:
        """Maps weight_after_surgery to aind model"""
        return self._parse_weight_str(self._nsb.weight_after_surgery)

    @property
    def aind_weight_before_surger(self) -> Optional[float]:
        """Maps weight_before_surger to aind model"""
        return self._parse_weight_str(self._nsb.weight_before_surger)

    @property
    def aind_work_station1st_injection(self) -> Optional[str]:
        """Maps work_station1st_injection to aind model"""
        return (
            None
            if self._nsb.work_station1st_injection is None
            else {
                self._nsb.work_station1st_injection.SELECT: None,
                self._nsb.work_station1st_injection.SWS_1: (
                    self._nsb.work_station1st_injection.SWS_1.value
                ),
                self._nsb.work_station1st_injection.SWS_2: (
                    self._nsb.work_station1st_injection.SWS_2.value
                ),
                self._nsb.work_station1st_injection.SWS_3: (
                    self._nsb.work_station1st_injection.SWS_3.value
                ),
                self._nsb.work_station1st_injection.SWS_4: (
                    self._nsb.work_station1st_injection.SWS_4.value
                ),
                self._nsb.work_station1st_injection.SWS_5: (
                    self._nsb.work_station1st_injection.SWS_5.value
                ),
                self._nsb.work_station1st_injection.SWS_6: (
                    self._nsb.work_station1st_injection.SWS_6.value
                ),
                self._nsb.work_station1st_injection.SWS_7: (
                    self._nsb.work_station1st_injection.SWS_7.value
                ),
                self._nsb.work_station1st_injection.SWS_8: (
                    self._nsb.work_station1st_injection.SWS_8.value
                ),
                self._nsb.work_station1st_injection.SWS_9: (
                    self._nsb.work_station1st_injection.SWS_9.value
                ),
            }.get(self._nsb.work_station1st_injection, None)
        )

    @property
    def aind_work_station2nd_injection(self) -> Optional[str]:
        """Maps work_station2nd_injection to aind model"""
        return (
            None
            if self._nsb.work_station2nd_injection is None
            else {
                self._nsb.work_station2nd_injection.SELECT: None,
                self._nsb.work_station2nd_injection.SWS_1: (
                    self._nsb.work_station2nd_injection.SWS_1.value
                ),
                self._nsb.work_station2nd_injection.SWS_2: (
                    self._nsb.work_station2nd_injection.SWS_2.value
                ),
                self._nsb.work_station2nd_injection.SWS_3: (
                    self._nsb.work_station2nd_injection.SWS_3.value
                ),
                self._nsb.work_station2nd_injection.SWS_4: (
                    self._nsb.work_station2nd_injection.SWS_4.value
                ),
                self._nsb.work_station2nd_injection.SWS_5: (
                    self._nsb.work_station2nd_injection.SWS_5.value
                ),
                self._nsb.work_station2nd_injection.SWS_6: (
                    self._nsb.work_station2nd_injection.SWS_6.value
                ),
                self._nsb.work_station2nd_injection.SWS_7: (
                    self._nsb.work_station2nd_injection.SWS_7.value
                ),
                self._nsb.work_station2nd_injection.SWS_8: (
                    self._nsb.work_station2nd_injection.SWS_8.value
                ),
                self._nsb.work_station2nd_injection.SWS_9: (
                    self._nsb.work_station2nd_injection.SWS_9.value
                ),
            }.get(self._nsb.work_station2nd_injection, None)
        )

    @property
    def aind_experimenter_full_name(self) -> str:
        """Map author id to experimenter name"""
        return (
            "NSB"
            if self.aind_author_id is None
            else f"NSB-{self.aind_author_id}"
        )

    @property
    def aind_anaesthetic_type(self) -> str:
        """Default anaesthetic type"""
        return "isoflurane"

    @property
    def aind_craniotomy_size(self) -> Optional[float]:
        """Map craniotomy type to size in mm"""
        return (
            None
            if self.aind_craniotomy_type is None
            else {
                self.aind_craniotomy_type.FIVE_MM: 5,
                self.aind_craniotomy_type.THREE_MM: 3,
            }.get(self.aind_craniotomy_type, None)
        )

    @property
    def aind_craniotomy_coordinates_reference(
        self,
    ) -> Optional[CoordinateReferenceLocation]:
        """Map craniotomy type to CoordinateReferenceLocation"""
        return (
            None
            if self.aind_craniotomy_type is None
            else {
                self.aind_craniotomy_type.VISCTX: (
                    CoordinateReferenceLocation.LAMBDA
                )
            }.get(self.aind_craniotomy_type, None)
        )

    @property
    def aind_inj1_coordinates_reference(
        self,
    ) -> Optional[CoordinateReferenceLocation]:
        """Map nsb inj1 field to CoordinateReferenceLocation"""
        if (
            self._nsb.virus_a_p is not None
            and "LAMBDA" in self._nsb.virus_a_p.upper()
        ):
            return CoordinateReferenceLocation.LAMBDA
        else:
            return None

    @property
    def aind_inj2_coordinates_reference(
        self,
    ) -> Optional[CoordinateReferenceLocation]:
        """Map nsb inj2 field to CoordinateReferenceLocation"""
        if (
            self._nsb.ap2nd_inj is not None
            and "LAMBDA" in self._nsb.ap2nd_inj.upper()
        ):
            return CoordinateReferenceLocation.LAMBDA
        else:
            return None

    def get_head_frame_procedure(self) -> Headframe:
        """Get head frame procedure"""
        return Headframe.construct(
            start_date=self.aind_date_of_surgery,
            end_date=self.aind_date_of_surgery,
            experimenter_full_name=self.aind_experimenter_full_name,
            iacuc_protocol=self.aind_iacuc_protocol,
            animal_weight_prior=self.aind_weight_before_surger,
            animal_weight_post=self.aind_weight_after_surgery,
            anaesthesia=Anaesthetic.construct(
                type=self.aind_anaesthetic_type,
                level=self.aind_hp_iso_level,
            ),
            headframe_type=self.aind_headpost_type.headframe_type,
            headframe_part_number=(
                self.aind_headpost_type.headframe_part_number
            ),
            well_part_number=self.aind_headpost_type.well_part_number,
            well_type=self.aind_headpost_type.well_type,
        )

    def get_craniotomy_procedure(self) -> Craniotomy:
        """Get craniotomy procedure"""
        return Craniotomy.construct(
            start_date=self.aind_date_of_surgery,
            end_date=self.aind_date_of_surgery,
            experimenter_full_name=self.aind_experimenter_full_name,
            iacuc_protocol=self.aind_iacuc_protocol,
            animal_weight_prior=self.aind_weight_before_surger,
            animal_weight_post=self.aind_weight_after_surgery,
            anaesthesia=Anaesthetic.construct(
                type=self.aind_anaesthetic_type,
                level=self.aind_hp_iso_level,
            ),
            craniotomy_type=self.aind_craniotomy_type,
            craniotomy_hemisphere=self.aind_hp_loc,
            craniotomy_coordinates_ml=self.aind_hp_m_l,
            craniotomy_coordinates_ap=self.aind_hp_a_p,
            craniotomy_coordinates_reference=(
                self.aind_craniotomy_coordinates_reference
            ),
            bregma_to_lambda_distance=self.aind_breg2_lamb,
            craniotomy_size=self.aind_craniotomy_size,
            dura_removed=self.aind_hp_durotomy,
            workstation_id=self.aind_hp_work_station,
        )

    def get_first_injection_procedure(self) -> BrainInjection:
        """Get first injection procedure"""
        if self.aind_inj1_type == InjectionType.NANOJECT:
            return NanojectInjection.construct(
                start_date=self.aind_date1st_injection,
                end_date=self.aind_date1st_injection,
                experimenter_full_name=self.aind_experimenter_full_name,
                animal_weight_prior=self.aind_first_injection_weight_be,
                animal_weight_post=self.aind_first_injection_weight_af,
                anaesthesia=Anaesthetic.construct(
                    type=self.aind_anaesthetic_type,
                    duration=self.aind_first_injection_iso_durat,
                    level=self.aind_hp_iso_level,
                ),
                injection_materials=(
                    [
                        InjectionMaterial.construct(
                            full_genome_name=self.aind_inj1_virus_strain_rt
                        )
                    ]
                ),
                recovery_time=self.aind_first_inj_recovery,
                injection_duration=self.aind_inj1_lenghtof_time,
                workstation_id=self.aind_work_station1st_injection,
                instrument_id=self.aind_nanoject_number_inj10,
                injection_coordinate_ml=self.aind_virus_m_l,
                injection_coordinate_ap=self.aind_virus_a_p,
                injection_coordinate_depth=self.aind_virus_d_v,
                injection_coordinate_reference=(
                    self.aind_inj1_coordinates_reference
                ),
                bregma_to_lambda_distance=self.aind_breg2_lamb,
                injection_angle=self.aind_inj1angle0,
                injection_hemisphere=self.aind_virus_hemisphere,
                injection_volume=self.aind_inj1_vol,
            )
        elif self.aind_inj1_type == InjectionType.IONTOPHORESIS:
            return IontophoresisInjection.construct(
                start_date=self.aind_date1st_injection,
                end_date=self.aind_date1st_injection,
                experimenter_full_name=self.aind_experimenter_full_name,
                animal_weight_prior=self.aind_first_injection_weight_be,
                animal_weight_post=self.aind_first_injection_weight_af,
                anaesthesia=Anaesthetic.construct(
                    type=self.aind_anaesthetic_type,
                    duration=self.aind_first_injection_iso_durat,
                    level=self.aind_hp_iso_level,
                ),
                injection_materials=(
                    [
                        InjectionMaterial.construct(
                            full_genome_name=self.aind_inj1_virus_strain_rt
                        )
                    ]
                ),
                recovery_time=self.aind_first_inj_recovery,
                injection_duration=self.aind_inj1_lenghtof_time,
                workstation_id=self.aind_work_station1st_injection,
                instrument_id=self.aind_ionto_number_inj1,
                injection_coordinate_ml=self.aind_virus_m_l,
                injection_coordinate_ap=self.aind_virus_a_p,
                injection_coordinate_depth=self.aind_virus_d_v,
                injection_coordinate_reference=(
                    self.aind_inj1_coordinates_reference
                ),
                bregma_to_lambda_distance=self.aind_breg2_lamb,
                injection_angle=self.aind_inj1angle0,
                injection_hemisphere=self.aind_virus_hemisphere,
                injection_current=self.aind_inj1_current,
                alternating_current=self.aind_inj1_alternating_time,
            )
        else:
            return BrainInjection.construct(
                start_date=self.aind_date1st_injection,
                end_date=self.aind_date1st_injection,
                experimenter_full_name=self.aind_experimenter_full_name,
                animal_weight_prior=self.aind_first_injection_weight_be,
                animal_weight_post=self.aind_first_injection_weight_af,
                anaesthesia=Anaesthetic.construct(
                    type=self.aind_anaesthetic_type,
                    duration=self.aind_first_injection_iso_durat,
                    level=self.aind_hp_iso_level,
                ),
                injection_materials=(
                    [
                        InjectionMaterial.construct(
                            full_genome_name=self.aind_inj1_virus_strain_rt
                        )
                    ]
                ),
                recovery_time=self.aind_first_inj_recovery,
                injection_duration=self.aind_inj1_lenghtof_time,
                workstation_id=self.aind_work_station1st_injection,
                injection_coordinate_ml=self.aind_virus_m_l,
                injection_coordinate_ap=self.aind_virus_a_p,
                injection_coordinate_depth=self.aind_virus_d_v,
                injection_coordinate_reference=(
                    self.aind_inj1_coordinates_reference
                ),
                bregma_to_lambda_distance=self.aind_breg2_lamb,
                injection_angle=self.aind_inj1angle0,
                injection_hemisphere=self.aind_virus_hemisphere,
            )

    def get_second_injection_procedure(self) -> BrainInjection:
        """Get second injection procedure"""
        if self.aind_inj2_type == InjectionType.NANOJECT:
            return NanojectInjection.construct(
                start_date=self.aind_date2nd_injection,
                end_date=self.aind_date2nd_injection,
                experimenter_full_name=self.aind_experimenter_full_name,
                animal_weight_prior=self.aind_second_injection_weight_b,
                animal_weight_post=self.aind_second_injection_weight_a,
                anaesthesia=Anaesthetic.construct(
                    type=self.aind_anaesthetic_type,
                    duration=self.aind_second_injection_iso_dura,
                    level=self.aind_hp_iso_level,
                ),
                injection_materials=(
                    [
                        InjectionMaterial.construct(
                            full_genome_name=self.aind_inj2_virus_strain_rt
                        )
                    ]
                ),
                recovery_time=self.aind_second_inj_recover,
                injection_duration=self.aind_inj2_lenghtof_time,
                workstation_id=self.aind_work_station2nd_injection,
                instrument_id=self.aind_nanoject_number_inj2,
                injection_coordinate_ml=self.aind_ml2nd_inj,
                injection_coordinate_ap=self.aind_ap2nd_inj,
                injection_coordinate_depth=self.aind_dv2nd_inj,
                injection_coordinate_reference=(
                    self.aind_inj2_coordinates_reference
                ),
                bregma_to_lambda_distance=self.aind_breg2_lamb,
                injection_angle=self.aind_inj2angle0,
                injection_hemisphere=self.aind_hemisphere2nd_inj,
                injection_volume=self.aind_inj2_vol,
            )
        elif self.aind_inj2_type == InjectionType.IONTOPHORESIS:
            return IontophoresisInjection.construct(
                start_date=self.aind_date2nd_injection,
                end_date=self.aind_date2nd_injection,
                experimenter_full_name=self.aind_experimenter_full_name,
                animal_weight_prior=self.aind_second_injection_weight_b,
                animal_weight_post=self.aind_second_injection_weight_a,
                anaesthesia=Anaesthetic.construct(
                    type=self.aind_anaesthetic_type,
                    duration=self.aind_second_injection_iso_dura,
                    level=self.aind_hp_iso_level,
                ),
                injection_materials=(
                    [
                        InjectionMaterial.construct(
                            full_genome_name=self.aind_inj2_virus_strain_rt
                        )
                    ]
                ),
                recovery_time=self.aind_second_inj_recover,
                injection_duration=self.aind_inj2_lenghtof_time,
                workstation_id=self.aind_work_station2nd_injection,
                instrument_id=self.aind_ionto_number_inj2,
                injection_coordinate_ml=self.aind_ml2nd_inj,
                injection_coordinate_ap=self.aind_ap2nd_inj,
                injection_coordinate_depth=self.aind_dv2nd_inj,
                injection_coordinate_reference=(
                    self.aind_inj2_coordinates_reference
                ),
                bregma_to_lambda_distance=self.aind_breg2_lamb,
                injection_angle=self.aind_inj2angle0,
                injection_hemisphere=self.aind_hemisphere2nd_inj,
                injection_current=self.aind_inj2_current,
                alternating_current=self.aind_inj2_alternating_time,
            )
        else:
            return BrainInjection.construct(
                start_date=self.aind_date2nd_injection,
                end_date=self.aind_date2nd_injection,
                experimenter_full_name=self.aind_experimenter_full_name,
                animal_weight_prior=self.aind_second_injection_weight_b,
                animal_weight_post=self.aind_second_injection_weight_a,
                anaesthesia=Anaesthetic.construct(
                    type=self.aind_anaesthetic_type,
                    duration=self.aind_second_injection_iso_dura,
                    level=self.aind_hp_iso_level,
                ),
                injection_materials=(
                    [
                        InjectionMaterial.construct(
                            full_genome_name=self.aind_inj2_virus_strain_rt
                        )
                    ]
                ),
                recovery_time=self.aind_second_inj_recover,
                injection_duration=self.aind_inj2_lenghtof_time,
                workstation_id=self.aind_work_station2nd_injection,
                injection_coordinate_ml=self.aind_ml2nd_inj,
                injection_coordinate_ap=self.aind_ap2nd_inj,
                injection_coordinate_depth=self.aind_dv2nd_inj,
                injection_coordinate_reference=(
                    self.aind_inj2_coordinates_reference
                ),
                bregma_to_lambda_distance=self.aind_breg2_lamb,
                injection_angle=self.aind_inj2angle0,
                injection_hemisphere=self.aind_hemisphere2nd_inj,
            )

    def get_fiber_implant(self):
        """Get a fiber implant procedure"""
        ophys_probes = []
        if self.aind_fiber_implant1:
            ophys_probes.append(
                OphysProbe.construct(
                    name=ProbeName.PROBE_A,
                    stereotactic_coordinate_ap=self.aind_virus_a_p,
                    stereotactic_coordinate_ml=self.aind_virus_m_l,
                    stereotactic_coordinate_dv=self.aind_fiber_implant1_dv,
                    stereotactic_coordinate_reference=(
                        self.aind_inj1_coordinates_reference
                    ),
                    bregma_to_lambda_distance=self.aind_breg2_lamb,
                    angle=self.aind_inj1_angle_v2,
                )
            )
        if self.aind_fiber_implant2:
            ophys_probes.append(
                OphysProbe.construct(
                    name=ProbeName.PROBE_B,
                    stereotactic_coordinate_ap=self.aind_ap2nd_inj,
                    stereotactic_coordinate_ml=self.aind_ml2nd_inj,
                    stereotactic_coordinate_dv=self.aind_fiber_implant2_dv,
                    stereotactic_coordinate_reference=(
                        self.aind_inj2_coordinates_reference
                    ),
                    bregma_to_lambda_distance=self.aind_breg2_lamb,
                    angle=self.aind_inj2_angle_v2,
                )
            )
        return FiberImplant.construct(
            start_date=self.aind_date_of_surgery,
            end_date=self.aind_date_of_surgery,
            experimenter_full_name=self.aind_experimenter_full_name,
            iacuc_protocol=self.aind_iacuc_protocol,
            animal_weight_prior=self.aind_weight_before_surger,
            animal_weight_post=self.aind_weight_after_surgery,
            probes=ophys_probes,
        )

    def get_basic_subject_procedure(self) -> SubjectProcedure:
        """Get a basic subject procedure"""
        return SubjectProcedure.construct(
            start_date=self.aind_date_of_surgery,
            end_date=self.aind_date_of_surgery,
            experimenter_full_name=self.aind_experimenter_full_name,
            iacuc_protocol=self.aind_iacuc_protocol,
            animal_weight_prior=self.aind_weight_before_surger,
            animal_weight_post=self.aind_weight_after_surgery,
        )

    @property
    def has_injection_procedure(self) -> bool:
        """Return true if injection procedure in nsb list"""
        if self._nsb.procedure is None:
            return False
        else:
            return {
                self._nsb.procedure.INJHPC: True,
                self._nsb.procedure.INJWHC_NP: True,
                self._nsb.procedure.HPINJ: True,
                self._nsb.procedure.HP_INJECTION_OPTIC_FIBER: True,
                self._nsb.procedure.STEREOTAXIC_INJECTION: True,
            }.get(self._nsb.procedure, False)

    @property
    def has_craniotomy_procedure(self) -> bool:
        """Return true if craniotomy procedure in nsb list"""
        if self._nsb.procedure is None:
            return False
        else:
            return {
                self._nsb.procedure.HPC_CAM: True,
                self._nsb.procedure.HPC_MULTISCOPE: True,
                self._nsb.procedure.HPC_NEUROPIXEL_STYLE: True,
                self._nsb.procedure.INJHPC: True,
                self._nsb.procedure.WHC_NP: True,
                self._nsb.procedure.INJWHC_NP: True,
            }.get(self._nsb.procedure, False)

    @property
    def has_head_frame_procedure(self) -> bool:
        """Return true if headframe procedure in nsb list"""
        if self._nsb.procedure is None:
            return False
        else:
            return {
                self._nsb.procedure.HPC_CAM: True,
                self._nsb.procedure.HPC_MULTISCOPE: True,
                self._nsb.procedure.HPC_NEUROPIXEL_STYLE: True,
                self._nsb.procedure.INJHPC: True,
                self._nsb.procedure.HPINJ: True,
                self._nsb.procedure.HP_TRANSCRANIAL_FOR_ISI: True,
                self._nsb.procedure.HP_ONLY: True,
                self._nsb.procedure.HP_INJECTION_OPTIC_FIBER: True,
            }.get(self._nsb.procedure, False)

    @property
    def has_fiber_implant_procedure(self) -> bool:
        """Return true if fiber implant procedure in nsb list"""
        if self._nsb.procedure is None:
            return False
        else:
            return {
                self._nsb.procedure.HP_INJECTION_OPTIC_FIBER: True,
            }.get(self._nsb.procedure, False)

    @property
    def has_unknown_procedures(self) -> bool:
        """Return true if no known procedures are found in nsb model"""
        if self._nsb.procedure is None:
            return True
        else:
            return not (
                self.has_injection_procedure
                or self.has_fiber_implant_procedure
                or self.has_craniotomy_procedure
                or self.has_head_frame_procedure
            )

    def get_procedures(self) -> List[SubjectProcedure]:
        """Get a list of subject procedures"""
        procedures = []
        if self.has_head_frame_procedure:
            procedures.append(self.get_head_frame_procedure())
        if self.has_injection_procedure:
            procedures.append(self.get_first_injection_procedure())
        if self.has_injection_procedure and self.aind_inj2_round is not None:
            procedures.append(self.get_second_injection_procedure())
        if self.has_craniotomy_procedure:
            procedures.append(self.get_craniotomy_procedure())
        if self.has_fiber_implant_procedure:
            procedures.append(self.get_fiber_implant())
        if self.has_unknown_procedures:
            procedures.append(self.get_basic_subject_procedure())
        return procedures



"""Module to create clients to connect to databases."""
from enum import Enum
from typing import List, Optional
from xml.etree import ElementTree as ET

import pyodbc
from aind_data_schema import Procedures, Subject
from aind_data_schema.procedures import (
    Perfusion,
    RetroOrbitalInjection,
    SubjectProcedure,
)
from aind_data_schema.subject import BackgroundStrain, Sex, Species
from fastapi.responses import JSONResponse

from aind_metadata_service.labtracks.query_builder import (
    LabTracksQueries,
    SubjectQueryColumns,
    TaskSetQueryColumns,
)
from aind_metadata_service.response_handler import Responses


class LabTracksClient:
    """This class contains the api to connect to LabTracks db."""

    def __init__(
        self,
        driver: str,
        server: str,
        port: str,
        db: str,
        user: str,
        password: str,
    ) -> None:
        """
        Initialize a client
        Parameters
        ----------
        driver : str
            ODBC Driver, example {FreeTDS}
        server : str
           Server DNS or IP
        port : int
           Server port
        db : str
           LabTracks Database name to connect to
        user : str
            LabTracks user
        password : str
            Password for LabTracks user
        """
        self.driver = driver
        self.server = server
        self.port = port
        self.db = db
        self.user = user
        self.password = password
        self.connection_str = (
            f"Driver={driver};"
            f"Server={server};"
            f"Port={port};"
            f"Database={db};"
            f"UID={user};"
            f"PWD={password};"
        )

    def create_session(self) -> pyodbc.Connection:
        """Use pyodbc to create a connection to sqlserver"""

        return pyodbc.connect(self.connection_str)

    @staticmethod
    def close_session(session: pyodbc.Connection) -> None:
        """Closes a pyodbc session connection"""
        session.close()

    def get_subject_info(self, subject_id: str) -> JSONResponse:
        """
        Method to retrieve subject from subject_id (int)
        Parameters
        ----------
        subject_id: str
        """
        try:
            query = LabTracksQueries.subject_from_subject_id(subject_id)
            session = self.create_session()
            cursor = session.cursor()
            cursor.execute(query)
            column_names = cursor.description
            columns = [column[0].lower() for column in column_names]
            fetched_rows = cursor.fetchall()
            cursor.close()
            self.close_session(session)
            results = []
            for row in fetched_rows:
                results.append(dict(zip(columns, row)))
            if not results:
                return Responses.no_data_found_response()
            else:
                lth = LabTracksResponseHandler()
                subjects = lth.map_response_to_subject(results)
                # Check if multiple unique items are returned
                if len(set([s.json() for s in subjects])) > 1:
                    response = Responses.multiple_items_found_response(
                        subjects
                    )
                    return response
                else:
                    subject = subjects[0]
                    response = Responses.model_response(subject)
                    return response
        except pyodbc.Error:
            return Responses.internal_server_error_response()

    def get_procedures_info(self, subject_id: str) -> JSONResponse:
        """
        Method to retrieve LAS procedures from subject_id (int)
        Parameters
        ----------
        subject_id: str
        """
        try:
            query = LabTracksQueries.procedures_from_subject_id(subject_id)
            session = self.create_session()
            cursor = session.cursor()
            cursor.execute(query)
            column_names = cursor.description
            columns = [column[0].lower() for column in column_names]
            fetched_rows = cursor.fetchall()
            cursor.close()
            self.close_session(session)
            results = []
            for row in fetched_rows:
                results.append(dict(zip(columns, row)))
            if not results:
                response = Responses.no_data_found_response()
            else:
                lth = LabTracksResponseHandler()
                subject_procedures = lth.map_response_to_procedures(results)
                procedures = Procedures.construct(subject_id=subject_id)
                procedures.subject_procedures = subject_procedures
                response = Responses.model_response(procedures)
            return response
        except pyodbc.Error:
            return Responses.internal_server_error_response()


class MouseCustomClassFields(Enum):
    """
    LabTracks stores an XML string in the class_values field. We can extract
    the full genotype of the mouse from that field.
    """

    RESERVED_BY = "Reserved_by"
    RESERVED_DATE = "Reserve_Date"
    REASON = "Reason"
    FULL_GENOTYPE = "Full_Genotype"


class LabTracksSex(Enum):
    """How LabTracks labels the sex of the species."""

    MALE = "M"
    FEMALE = "F"


class LabTracksSpecies(Enum):
    """How LabTracks stores the name of the species."""

    MOUSE = "mouse"
    RAT = "rat"


class LabTracksBgStrain(Enum):
    """How LabTracks labels its strains"""

    # TODO: Double check this
    C57BL_6J = "C57BL/6J"
    BALB_C = "BALB/c"


class LabTracksProcedures(Enum):
    """Keywords for LabTracks procedures"""

    PERFUSION = "Perfusion"
    RO_INJECTION = "RO Injection"


class LabTracksResponseHandler:
    """This class will contain methods to handle the response from LabTracks"""

    @staticmethod
    def _map_class_values_to_genotype(
        class_values: Optional[str],
    ) -> Optional[str]:
        """
        Extracts the full genotype from the class values field.
        Parameters
        ----------
        class_values : Optional[str]
          XML string that is expected to have a Full_Genotype field.

        Returns
        -------
          The text inside the Full_Genotype field or None if the XML string
          does not contain the Full_Genotype.

        """
        if class_values is None:
            return None
        else:
            xml_root = ET.fromstring(class_values)
            full_genotype = xml_root.find(
                MouseCustomClassFields.FULL_GENOTYPE.value
            )
            if full_genotype is not None:
                full_genotype_text = full_genotype.text
                return full_genotype_text
            else:
                return None

    @staticmethod
    def _map_species(species: Optional[str]) -> Optional[Species]:
        """
        Maps the LabTracks species to the aind_data_schema.subject.Species
        Parameters
        ----------
        species : str
          LabTracks species name

        Returns
        -------
          An aind_data_schema.subject.Species or None if no mapping exists.

        """
        if species is None:
            return None
        if species.lower() == LabTracksSpecies.MOUSE.value.lower():
            return Species.MUS_MUSCULUS
        else:
            return None

    @staticmethod
    def _map_sex(sex: Optional[str]) -> Optional[Sex]:
        """
        Maps the LabTracks Sex enum to the aind_data_schema.subject.Sex
        Parameters
        ----------
        sex : Optional[str]
          LabTracks sex name

        Returns
        -------
          An aind_data_schema.subject.Sex or None if no such mapping exists.

        """
        if sex is None:
            return None
        if sex.lower() == LabTracksSex.MALE.value.lower():
            return Sex.MALE
        elif sex.lower() == LabTracksSex.FEMALE.value.lower():
            return Sex.FEMALE
        else:
            return None

    @staticmethod
    def _map_to_background_strain(
        bg_strain: Optional[str],
    ) -> Optional[BackgroundStrain]:
        """
        Maps the LabTracks BG Strain enum to the
        aind_data_schema.subject.BackgroundStrain
        Parameters
        ----------
        bg_strain : Optional[str]

        Returns
        -------
        Optional[BackgroundStrain]
        """

        if bg_strain is None:
            return None
        if bg_strain.lower() == LabTracksBgStrain.C57BL_6J.value.lower():
            return BackgroundStrain.C57BL_6J
        elif bg_strain.lower() == LabTracksBgStrain.BALB_C.value.lower():
            return BackgroundStrain.BALB_c
        else:
            return None

    def map_response_to_subject(self, results: List[dict]) -> List[Subject]:
        """
        Maps a response from LabTracks to an aind_data_schema.Subject
        Parameters
        ----------
        results : List[dict]
          Results pulled from LabTracks. In the form of [row]

        Returns
        -------
          A List of mapped Subjects (not validated)

        """

        subjects = []

        for result in results:
            class_values = result.get(SubjectQueryColumns.CLASS_VALUES.value)
            full_genotype = self._map_class_values_to_genotype(class_values)
            sex = self._map_sex(result.get(SubjectQueryColumns.SEX.value))
            species = self._map_species(
                result.get(SubjectQueryColumns.SPECIES_NAME.value)
            )
            paternal_genotype = self._map_class_values_to_genotype(
                result.get(SubjectQueryColumns.PATERNAL_CLASS_VALUES.value)
            )
            maternal_genotype = self._map_class_values_to_genotype(
                result.get(SubjectQueryColumns.MATERNAL_CLASS_VALUES.value)
            )
            paternal_id = result.get(SubjectQueryColumns.PATERNAL_ID.value)
            paternal_id_str = str(paternal_id) if paternal_id else None
            maternal_id = result.get(SubjectQueryColumns.MATERNAL_ID.value)
            maternal_id_str = str(maternal_id) if maternal_id else None
            subject_id = result.get(SubjectQueryColumns.ID.value)
            subject_id_str = str(subject_id) if subject_id else None
            datetime_of_birth = result.get(
                SubjectQueryColumns.BIRTH_DATE.value
            )
            date_of_birth = (
                datetime_of_birth.date() if datetime_of_birth else None
            )
            breeding_group = result.get(SubjectQueryColumns.GROUP_NAME.value)
            background_strain = self._map_to_background_strain(
                result.get(SubjectQueryColumns.GROUP_DESCRIPTION.value)
            )
            subject = Subject.construct(
                subject_id=subject_id_str,
                species=species,
                paternal_genotype=paternal_genotype,
                paternal_id=paternal_id_str,
                maternal_genotype=maternal_genotype,
                maternal_id=maternal_id_str,
                sex=sex,
                date_of_birth=date_of_birth,
                genotype=full_genotype,
                breeding_group=breeding_group,
                background_strain=background_strain,
            )
            subjects.append(subject)
        return subjects

    @staticmethod
    def map_response_to_procedures(
        results: List[dict],
    ) -> List[SubjectProcedure]:
        """
        Maps a response from LabTracks to an aind_data_schema.Procedure
        Parameters
        ----------
        results : List[dict]
            Results pulled from LabTracks. In the form of [row]

        Returns
        -------
            A List of mapped Subjects (not validated)
        """

        procedures_list = []

        for result in results:
            start_date = result.get(TaskSetQueryColumns.DATE_START.value)
            if start_date:
                start_date = start_date.date()
            end_date = result.get(TaskSetQueryColumns.DATE_END.value)
            if end_date:
                end_date = end_date.date()
            experimenter_full_name = result.get(
                TaskSetQueryColumns.INVESTIGATOR_ID.value
            )
            iacuc_protocol = result.get(
                TaskSetQueryColumns.PROTOCOL_NUMBER.value
            )
            type_name = result.get(TaskSetQueryColumns.TYPE_NAME.value)
            if type_name:
                if LabTracksProcedures.PERFUSION.value in type_name:
                    output_specimen_ids = [
                        result.get(TaskSetQueryColumns.TASK_OBJECT.value)
                    ]
                    perfusion = Perfusion.construct(
                        start_date=start_date,
                        end_date=end_date,
                        experimenter_full_name=experimenter_full_name,
                        iacuc_protocol=iacuc_protocol,
                        output_specimen_ids=output_specimen_ids,
                    )
                    procedures_list.append(perfusion)

                elif LabTracksProcedures.RO_INJECTION.value in type_name:
                    # TODO: parse inj info from comments
                    ro_injection = RetroOrbitalInjection.construct(
                        start_date=start_date,
                        end_date=end_date,
                        experimenter_full_name=experimenter_full_name,
                        iacuc_protocol=iacuc_protocol,
                    )
                    procedures_list.append(ro_injection)
        return procedures_list



"""Module for client library."""
from enum import Enum

import requests


class StatusCodes(Enum):
    """Enum class of status codes"""

    CONNECTION_ERROR = 503
    INTERNAL_SERVER_ERROR = 500
    MULTIPLE_RESPONSES = 300
    VALID_DATA = 200
    INVALID_DATA = 406
    NO_DATA_FOUND = 404
    MULTI_STATUS = 207


class AindMetadataServiceClient:
    """Class to handle client api calls to the service."""

    def __init__(self, domain: str):
        """
        Class constructor
        Parameters
        ----------
        domain : str
          url of the domain
        """

        self.domain = domain
        self.subject_url = f"{self.domain}/subject"
        self.procedures_url = f"{self.domain}/procedures"

    def get_subject(self, subject_id: str) -> requests.Response:
        """
        Retrieve a subject response from the server
        Parameters
        ----------
        subject_id : str
          id of the subject

        Returns
        -------
        requests.Response

        """
        url = "/".join([self.subject_url, subject_id])
        with requests.get(url) as response:
            return response

    def get_procedures(self, subject_id: str) -> requests.Response:
        """
        Retrieve a procedures response from the server
        Parameters
        ----------
        subject_id : str
          id of the subject

        Returns
        -------
        requests.Response

        """
        url = "/".join([self.procedures_url, subject_id])
        with requests.get(url) as response:
            return response



"""Tests NSB 2023 data model is parsed correctly"""

import json
import logging
import os
from copy import deepcopy
from pathlib import Path
from typing import List, Tuple
from unittest import TestCase
from unittest import main as unittest_main

from aind_data_schema.procedures import BrainInjection, CraniotomyType

from aind_metadata_service.sharepoint.nsb2023.mapping import (
    BurrHoleInfo,
    HeadPost,
    HeadPostInfo,
    HeadPostType,
    MappedNSBList,
)
from aind_metadata_service.sharepoint.nsb2023.models import NSBList

if os.getenv("LOG_LEVEL"):  # pragma: no cover
    logging.basicConfig(level=os.getenv("LOG_LEVEL"))

TEST_DIR = Path(os.path.dirname(os.path.realpath(__file__))) / ".." / ".."
DIR_RAW = TEST_DIR / "resources" / "sharepoint" / "nsb2023" / "raw"
DIR_MAP = TEST_DIR / "resources" / "sharepoint" / "nsb2023" / "mapped"
LIST_ITEM_FILE_NAMES = os.listdir(DIR_RAW)
sorted(LIST_ITEM_FILE_NAMES)
LIST_ITEM_FILE_PATHS = [DIR_RAW / str(f) for f in LIST_ITEM_FILE_NAMES]
MAPPED_ITEM_FILE_NAMES = os.listdir(DIR_MAP)
sorted(MAPPED_ITEM_FILE_NAMES)
MAPPED_FILE_PATHS = [DIR_MAP / str(f) for f in MAPPED_ITEM_FILE_NAMES]


class TestNSB2023Parsers(TestCase):
    """Tests methods in NSB2023Mapping class"""

    @classmethod
    def setUpClass(cls):
        """Load json files before running tests."""
        cls.list_items = cls._load_json_files()

    @staticmethod
    def _load_json_files() -> List[Tuple[dict, dict, str]]:
        """Reads raw data and expected data into json"""
        list_items = []
        for file_path in LIST_ITEM_FILE_PATHS:
            mapped_file_path = (
                file_path.parent.parent
                / "mapped"
                / ("mapped_" + file_path.name)
            )
            with open(file_path) as f:
                contents = json.load(f)
            with open(mapped_file_path) as f:
                mapped_contents = json.load(f)
            list_items.append((contents, mapped_contents, file_path.name))
            list_items.sort(key=lambda x: x[2])
        return list_items

    def test_parser(self):
        """Checks that raw data is parsed correctly"""
        for list_item in self.list_items:
            raw_data = list_item[0]
            expected_mapped_data = list(list_item[1])
            expected_mapped_data.sort(key=lambda x: str(x))
            raw_file_name = list_item[2]
            logging.debug(f"Processing file: {raw_file_name}")
            nsb_model = NSBList.parse_obj(raw_data)
            mapper = MappedNSBList(nsb=nsb_model)
            mapped_procedures = mapper.get_procedures()
            mapped_procedure_json = [
                json.loads(p.json()) for p in mapped_procedures
            ]
            mapped_procedure_json.sort(key=lambda x: str(x))
            self.assertEqual(expected_mapped_data, mapped_procedure_json)

    def test_properties(self):
        """Tests that the properties are parsed correctly."""
        list_item = self.list_items[0]
        raw_data = deepcopy(list_item[0])
        nsb_model = NSBList.parse_obj(raw_data)
        mapped_model = MappedNSBList(nsb=nsb_model)
        props = []
        for k in dir(mapped_model.__class__):
            cls = mapped_model.__class__
            attr = getattr(cls, k)
            if isinstance(attr, property):
                props.append(getattr(mapped_model, k))
        self.assertEqual(144, len(props))

    def test_parse_basic_float_str(self):
        """Tests parsing of basic float strings"""
        self.assertEqual(MappedNSBList._parse_basic_float_str("0.25"), 0.25)
        self.assertIsNone(MappedNSBList._parse_basic_float_str("abc"))

    def test_craniotomy_edge_case(self):
        """Tests other craniotomy cases"""

        # Check WHC type
        list_item = self.list_items[2]
        raw_data = deepcopy(list_item[0])
        raw_data["Procedure"] = "WHC NP"
        raw_data["CraniotomyType"] = "WHC"
        nsb_model1 = NSBList.parse_obj(raw_data)
        mapper = MappedNSBList(nsb=nsb_model1)
        mapped_procedure1 = mapper.get_procedures()
        mapped_procedure1.sort(key=lambda x: str(x))
        cran_proc1 = mapped_procedure1[0]
        self.assertEqual(
            CraniotomyType.WHC, getattr(cran_proc1, "craniotomy_type")
        )

        # Check OTHER type
        raw_data["Procedure"] = "WHC NP"
        raw_data["CraniotomyType"] = "Select..."
        nsb_model2 = NSBList.parse_obj(raw_data)
        mapper = MappedNSBList(nsb=nsb_model2)
        mapped_procedure2 = mapper.get_procedures()
        mapped_procedure2.sort(key=lambda x: str(x))
        cran_proc2 = mapped_procedure2[0]
        self.assertIsNone(getattr(cran_proc2, "craniotomy_type"))

    def test_headpost_edge_case(self):
        """Test edge case for HeadPostInfo class constructor"""
        hp_info = HeadPostInfo.from_hp_and_hp_type(
            hp=HeadPost.VISUAL_CTX, hp_type=HeadPostType.MESOSCOPE
        )
        exp_hp_info = HeadPostInfo(
            headframe_type=HeadPost.VISUAL_CTX.value,
            headframe_part_number="0160-100-10",
            well_type=HeadPostType.MESOSCOPE.value,
            well_part_number="0160-200-20",
        )
        self.assertEqual(exp_hp_info, hp_info)

    def test_injection_edge_case(self):
        """Tests the case where there is an INJ procedure, but the inj types
        are malformed. It should create generic BrainInjection objects."""
        list_item = self.list_items[3]
        raw_data = deepcopy(list_item[0])
        raw_data["Inj1Type"] = "Select..."
        nsb_model = NSBList.parse_obj(raw_data)
        mapper = MappedNSBList(nsb=nsb_model)
        mapped_procedure = mapper.get_procedures()
        proc_types = [type(p) for p in mapped_procedure]
        self.assertTrue(BrainInjection in proc_types)

    def test_burr_hole_to_probe_edge_case(self):
        """Tests edge case where burr hole number greater than 4"""
        list_item = self.list_items[2]
        raw_data = deepcopy(list_item[0])
        nsb_model = NSBList.parse_obj(raw_data)
        mapper = MappedNSBList(nsb=nsb_model)
        self.assertIsNone(mapper._map_burr_hole_number_to_probe(5))
        self.assertEqual(BurrHoleInfo(), mapper.burr_hole_info(5))


if __name__ == "__main__":
    unittest_main()



"""This module was autogenerated from sharepoint list schema. Please be
careful modifying the contents."""

from datetime import datetime
from enum import Enum, EnumMeta
from typing import Optional

from pydantic import BaseModel, Extra, Field


class OptionalEnumMeta(EnumMeta):
    """Return None if constructing from a value not in members."""

    def __call__(cls, value, *args, **kw):
        """Return none if value not in members. Else return enum."""
        if value in [m.value for m in cls.__members__.values()]:
            return super().__call__(value, *args, **kw)
        else:
            return None


class After1StInjection(Enum, metaclass=OptionalEnumMeta):
    """Enum class for After1StInjection"""

    BLEEDING = "Bleeding"
    BLEEDING_FLUID_COMING_UP = "Bleeding/Fluid coming up"
    FLUID_COMING_UP = "Fluid coming up"
    NA = "N/A"
    NO_ISSUES = "No Issues"
    SELECT = "Select..."
    SWELLING = "Swelling"


class After2NdInjection(Enum, metaclass=OptionalEnumMeta):
    """Enum class for After2NdInjection"""

    BLEEDING = "Bleeding"
    BLEEDING_FLUID_COMING_UP = "Bleeding/Fluid coming up"
    FLUID_COMING_UP = "Fluid coming up"
    NA = "N/A"
    NO_ISSUES = "No Issues"
    SELECT = "Select..."
    SWELLING = "Swelling"


class Contusion(Enum, metaclass=OptionalEnumMeta):
    """Enum class for Contusion"""

    MILD = "Mild"
    MODERATE = "Moderate"
    NA = "N/A"
    NONE = "None"
    SELECT = "Select..."
    SEVERE = "Severe"


class CraniotomyType(Enum, metaclass=OptionalEnumMeta):
    """Enum class for CraniotomyType"""

    FRONTAL_WINDOW_3MM = "Frontal Window 3mm"
    SELECT = "Select..."
    VISUAL_CORTEX_5MM = "Visual Cortex 5mm"
    WHC_2_P = "WHC 2P"
    WHC_NP = "WHC NP"


class During1StInjection(Enum, metaclass=OptionalEnumMeta):
    """Enum class for During1StInjection"""

    BLEEDING = "Bleeding"
    BLEEDING_FLUID_COMING = "Bleeding/Fluid coming"
    FLUID_COMING_UP = "Fluid coming up"
    NA = "N/A"
    NO_ISSUES = "No Issues"
    PIPETTE_CLOGGEDBROKE_TIP = "Pipette clogged/broke tip"
    SELECT = "Select..."
    SWELLING = "Swelling"


class During2NdInjection(Enum, metaclass=OptionalEnumMeta):
    """Enum class for During2NdInjection"""

    BLEEDING = "Bleeding"
    BLEEDING_FLUID_COMING_UP = "Bleeding/Fluid coming up"
    FLUID_COMING_UP = "Fluid coming up"
    NA = "N/A"
    NO_ISSUES = "No Issues"
    PIPETTE_CLOGGEDBROKE_TIP = "Pipette clogged/broke tip"
    SELECT = "Select..."
    SWELLING = "Swelling"


class Durotomy(Enum, metaclass=OptionalEnumMeta):
    """Enum class for Durotomy"""

    COMPLETE = "Complete"
    NA = "N/A"
    NO = "No"
    PARTIAL = "Partial"
    SELECT = "Select..."
    TORN_COMPLETE = "Torn & Complete"
    UNINTENTIONAL = "Unintentional"


class EdemaSwelling(Enum, metaclass=OptionalEnumMeta):
    """Enum class for EdemaSwelling"""

    MILD = "Mild"
    MODERATE = "Moderate"
    NA = "N/A"
    NONE = "None"
    SELECT = "Select..."
    SEVERE = "Severe"


class ExudateSeverity(Enum, metaclass=OptionalEnumMeta):
    """Enum class for ExudateSeverity"""

    MILD = "Mild"
    MODERATE = "Moderate"
    NONE = "None"
    SELECT = "Select..."
    SEVERE = "Severe"


class EyeAffected(Enum, metaclass=OptionalEnumMeta):
    """Enum class for EyeAffected"""

    BOTH = "Both"
    LEFT = "Left"
    NEITHER = "Neither"
    RIGHT = "Right"
    SELECT = "Select..."


class EyeIssue(Enum, metaclass=OptionalEnumMeta):
    """Enum class for EyeIssue"""

    CLOUD_POSSIBLY_BLIND = "Cloud/Possibly Blind"
    NONE = "None"
    OTHER_SEE_COMMENTS = "Other...(See comments)"
    PUFFY_SWOLLEN = "Puffy/Swollen"
    SELECT = "Select..."
    SLIGHTLY_CLOSED = "Slightly Closed"
    WEEPY = "Weepy"


class FirstRoundInjIsoLevel(Enum, metaclass=OptionalEnumMeta):
    """Enum class for FirstRoundInjIsoLevel"""

    N_025 = "0.25"
    N_050 = "0.50"
    N_075 = "0.75"
    N_100 = "1.00"
    N_125 = "1.25"
    N_150 = "1.50"
    N_175 = "1.75"
    N_200 = "2.00"
    N_225 = "2.25"
    N_250 = "2.50"
    N_275 = "2.75"
    N_300 = "3.00+"
    SELECT = "Select..."


class FirstRoundInjWorkstation(Enum, metaclass=OptionalEnumMeta):
    """Enum class for FirstRoundInjWorkstation"""

    SELECT = "Select..."
    SWS_1 = "SWS 1"
    SWS_2 = "SWS 2"
    SWS_3 = "SWS 3"
    SWS_4 = "SWS 4"
    SWS_5 = "SWS 5"
    SWS_6 = "SWS 6"
    SWS_7 = "SWS 7"
    SWS_8 = "SWS 8"
    SWS_9 = "SWS 9"


class Firstroundiontoissue(Enum, metaclass=OptionalEnumMeta):
    """Enum class for Firstroundiontoissue"""

    NA = "N/A"
    NO = "No"
    SELECT = "Select..."
    YES = "Yes"


class Firstroundiontonumber(Enum, metaclass=OptionalEnumMeta):
    """Enum class for Firstroundiontonumber"""

    IONTO_1 = "Ionto #1"
    IONTO_10 = "Ionto #10"
    IONTO_2 = "Ionto #2"
    IONTO_3 = "Ionto #3"
    IONTO_4 = "Ionto #4"
    IONTO_5 = "Ionto #5"
    IONTO_6 = "Ionto #6"
    IONTO_7 = "Ionto #7"
    IONTO_8 = "Ionto #8"
    IONTO_9 = "Ionto #9"
    NA = "N/A"
    SELECT = "Select..."


class Firstroundnanojectnumber(Enum, metaclass=OptionalEnumMeta):
    """Enum class for Firstroundnanojectnumber"""

    NA = "N/A"
    NJ1 = "NJ#1"
    NJ2 = "NJ#2"
    NJ3 = "NJ#3"
    NJ4 = "NJ#4"
    NJ5 = "NJ#5"
    NJ6 = "NJ#6"
    NJ7 = "NJ#7"
    NJ8 = "NJ#8"
    SELECT = "Select..."


class HeadpostType(Enum, metaclass=OptionalEnumMeta):
    """Enum class for HeadpostType"""

    AI_STRAIGHT_HEADBAR = "AI Straight Headbar"
    CAMSTYLE_HEADFRAME_016010 = "CAM-style headframe (0160-100-10 Rev A)"
    MESOSCOPESTYLE_WELL_WITH = (
        "Mesoscope-style well with NGC-style headframe "
        "(0160-200-20/0160-100-10)"
    )
    NEUROPIXELSTYLE_HEADFRAME = (
        "Neuropixel-style headframe (0160-100-10/0160-200-36)"
    )
    NGCSTYLE_HEADFRAME_NO_WEL = "NGC-style headframe, no well (0160-100-10)"
    SELECT = "Select..."
    WHC_42_WITH_NEUROPIXEL_WE = "WHC #42 with Neuropixel well and well cap"


class Hematoma(Enum, metaclass=OptionalEnumMeta):
    """Enum class for Hematoma"""

    MILD = "Mild"
    MODERATE = "Moderate"
    NA = "N/A"
    NONE = "None"
    SELECT = "Select..."
    SEVERE = "Severe"


class HpDurotomy(Enum, metaclass=OptionalEnumMeta):
    """Enum class for HpDurotomy"""

    NO = "No"
    SELECT = "Select..."
    YES = "Yes"


class HpIsoLevel(Enum, metaclass=OptionalEnumMeta):
    """Enum class for HpIsoLevel"""

    N_025 = "0.25"
    N_050 = "0.50"
    N_075 = "0.75"
    N_100 = "1.00"
    N_125 = "1.25"
    N_15 = "1.5"
    N_175 = "1.75"
    N_200 = "2.00"
    N_225 = "2.25"
    N_250 = "2.50"
    N_275 = "2.75"
    N_300 = "3.00+"
    SELECT = "Select..."


class HpLocation(Enum, metaclass=OptionalEnumMeta):
    """Enum class for HpLocation"""

    CENTER = "Center"
    LEFT = "Left"
    RIGHT = "Right"
    SELECT = "Select..."


class HpPerferations(Enum, metaclass=OptionalEnumMeta):
    """Enum class for HpPerferations"""

    LEFT = "Left"
    NA = "NA"
    RIGHT = "Right"
    SELECT_IF_APPLICABLE = "Select if applicable..."


class HpWorkStation(Enum, metaclass=OptionalEnumMeta):
    """Enum class for HpWorkStation"""

    SELECT = "Select..."
    SWS_1 = "SWS 1"
    SWS_2 = "SWS 2"
    SWS_3 = "SWS 3"
    SWS_4 = "SWS 4"
    SWS_5 = "SWS 5"
    SWS_6 = "SWS 6"
    SWS_7 = "SWS 7"
    SWS_8 = "SWS 8"
    SWS_9 = "SWS 9"


class IacucProtocol(Enum, metaclass=OptionalEnumMeta):
    """Enum class for IacucProtocol"""

    N_2001 = "2001"
    N_2002 = "2002"
    N_2003 = "2003"
    N_2004 = "2004"
    N_2005 = "2005"
    N_2006 = "2006"
    N_2011 = "2011"
    N_2102 = "2102"
    N_2103 = "2103"
    N_2104 = "2104"
    N_2105 = "2105"
    N_2106 = "2106"
    N_2107 = "2107"
    N_2108 = "2108"
    N_2109 = "2109"
    N_2110 = "2110"
    N_2113 = "2113"
    N_2115 = "2115"
    N_2117 = "2117"
    N_2201 = "2201"
    N_2202 = "2202"
    N_2205 = "2205"
    N_2212 = "2212"
    SELECT = "Select..."


class ImplantIdCoverslipType(Enum, metaclass=OptionalEnumMeta):
    """Enum class for ImplantIdCoverslipType"""

    CUSTOM_IMPLANT_ADD_DETAIL = (
        "Custom Implant (add details in comments section)"
    )
    N_2001 = "2001"
    N_2002 = "2002"
    N_3001 = "3001"
    N_5MM_STACKED_COVERSLIP = "5mm stacked coverslip"
    N_5MM_STACKED_COVERSLIP_W = (
        "5mm stacked coverslip (with silicone coating for Neuropixel)"
    )
    SELECT = "Select..."


class Inj1Angle(Enum, metaclass=OptionalEnumMeta):
    """Enum class for Inj1Angle"""

    N_0_DEGREES = "0 degrees"
    N_10_DEGREES = "10 degrees"
    N_15_DEGREES = "15 degrees"
    N_20_DEGREES = "20 degrees"
    N_30_DEGREES = "30 degrees"
    N_40_DEGREES = "40 degrees"
    SELECT = "Select..."


class Inj1Hemisphere(Enum, metaclass=OptionalEnumMeta):
    """Enum class for Inj1Hemisphere"""

    LEFT = "Left"
    RIGHT = "Right"
    SELECT = "Select..."


class Inj1Retsetting(Enum, metaclass=OptionalEnumMeta):
    """Enum class for Inj1Retsetting"""

    OFF = "Off"
    ON = "On"


class Inj1Round(Enum, metaclass=OptionalEnumMeta):
    """Enum class for Inj1Round"""

    NA = "NA"
    N_1ST = "1st"
    N_2ND = "2nd"
    SELECT = "Select..."


class Inj1Type(Enum, metaclass=OptionalEnumMeta):
    """Enum class for Inj1Type"""

    IONTOPHORESIS = "Iontophoresis"
    NANOJECT_PRESSURE = "Nanoject (Pressure)"
    SELECT = "Select..."


class Inj2Angle(Enum, metaclass=OptionalEnumMeta):
    """Enum class for Inj2Angle"""

    N_0_DEGREES = "0 degrees"
    N_10_DEGREES = "10 degrees"
    N_15_DEGREES = "15 degrees"
    N_20_DEGREES = "20 degrees"
    N_30_DEGREES = "30 degrees"
    N_40_DEGREES = "40 degrees"
    SELECT = "Select..."


class Inj2Hemisphere(Enum, metaclass=OptionalEnumMeta):
    """Enum class for Inj2Hemisphere"""

    LEFT = "Left"
    RIGHT = "Right"
    SELECT = "Select..."


class Inj2Retsetting(Enum, metaclass=OptionalEnumMeta):
    """Enum class for Inj2Retsetting"""

    OFF = "Off"
    ON = "On"


class Inj2Round(Enum, metaclass=OptionalEnumMeta):
    """Enum class for Inj2Round"""

    NA = "NA"
    N_1ST = "1st"
    N_2ND = "2nd"
    SELECT = "Select..."


class Inj2Type(Enum, metaclass=OptionalEnumMeta):
    """Enum class for Inj2Type"""

    IONTOPHORESIS = "Iontophoresis"
    NANOJECT_PRESSURE = "Nanoject (Pressure)"
    SELECT = "Select..."


class Iontonumberhpinj(Enum, metaclass=OptionalEnumMeta):
    """Enum class for Iontonumberhpinj"""

    IONTO_1 = "Ionto #1"
    IONTO_2 = "Ionto #2"
    IONTO_3 = "Ionto #3"
    IONTO_4 = "Ionto #4"
    IONTO_5 = "Ionto #5"
    IONTO_6 = "Ionto #6"
    IONTO_7 = "Ionto #7"
    IONTO_8 = "Ionto #8"
    IONTO_9 = "Ionto #9"


class Issue(Enum, metaclass=OptionalEnumMeta):
    """Enum class for Issue"""

    EXUDATE = "Exudate"
    EYE_ISSUE = "Eye Issue"
    METABOND = "Metabond"
    MULTIPLE_ISSUES_NOTES_IN = "Multiple Issues (Notes In comments)"
    NON_SURGICAL_ISSUE = "Non-Surgical Issue"
    OTHERWRITE_IN_COMMENTS = "Other..(write in comments)"
    REPLACEMENT_COVERSLIP = "Replacement: Coverslip"
    REPLACEMENT_KWIKCAST = "Replacement: Kwikcast"
    REPLACEMENT_ORING = "Replacement: O-ring"
    SCABBING = "Scabbing"
    SELECT = "Select..."
    Z_MOTION = "Z-Motion"


class Laceration(Enum, metaclass=OptionalEnumMeta):
    """Enum class for Laceration"""

    MILD = "Mild"
    MODERATE = "Moderate"
    NA = "N/A"
    NONE = "None"
    SELECT = "Select..."
    SEVERE = "Severe"


class LightCycle(Enum, metaclass=OptionalEnumMeta):
    """Enum class for LightCycle"""

    NORMAL_6AM_TO_8PM = "Normal (6am to 8pm)"
    REVERSE_9PM_TO_9AM = "Reverse (9pm to 9am)"
    SELECT = "Select..."


class LimsProjectCode(Enum, metaclass=OptionalEnumMeta):
    """Enum class for LimsProjectCode"""

    AINDDISCOVERY = "aind-discovery"
    AINDEPHYS = "aind-ephys"
    AINDMSMA = "aind-msma"
    AINDOPHYS = "aind-ophys"
    BRAINTV_VIRAL_STRATEGIES = "BraintvViralStrategies"
    CELLTYPES_TRANSGENIC_CHAR = "CelltypesTransgenicCharacterizationGCaMP"
    CITRICACIDPILOT = "Citricacidpilot"
    DYNAMIC_ROUTING_BEHAVIOR = "DynamicRoutingBehaviorDev"
    DYNAMIC_ROUTING_SURGICAL = "DynamicRoutingSurgicalDevelopment"
    DYNAMIC_ROUTING_TASK1_PRO = "DynamicRoutingTask1Production"
    DYNAMIC_ROUTING_ULTRA_OPT = "DynamicRoutingUltraOptotaggingBehavior"
    IS_IX = "ISIx"
    LEARNINGM_FISH_DEVELOPMEN = "LearningmFISHDevelopment"
    LEARNINGM_FISH_TASK1_A = "LearningmFISHTask1A"
    MINDSCOPE_TRANSGENIC_CHAR = "MindscopeTransgenicCharacterizationGCaMP"
    MOUSE_BRAIN_CELL_ATLAS_TR = "MouseBrainCellAtlasTranssynaptic"
    MOUSE_GENETIC_TOOLS_PROJE = "MouseGeneticToolsProjectionMapping"
    M_IVSCCMET = "mIVSCC-MET"
    M_IVSCCME_TX = "mIVSCC-METx"
    M_MPATC_HX = "mMPATCHx"
    M_M_PATCH = "mMPatch"
    OM_FISHRO_BINJECTIONVIRUS = "omFISHROBinjectionviruspilot"
    OM_FISH_CUX2_MESO = "omFISHCux2Meso"
    OM_FISH_GAD2_MESO = "omFISHGad2Meso"
    OM_FISH_GAD2_PILOT = "omFISHGad2Pilot"
    OM_FISH_RORB_PILOT = "omFISHRorbPilot"
    OM_FIS_HCOREGISTRATIONPIL = "omFISHcoregistrationpilot"
    OPENSCOPE_DEVELOPMENT = "OpenscopeDevelopment"
    OPEN_SCOPE_DENDRITE_COUPL = "OpenScopeDendriteCoupling"
    OPEN_SCOPE_GLOBAL_LOCAL_O = "OpenScopeGlobalLocalOddball"
    OPEN_SCOPE_ILLUSION = "OpenScopeIllusion"
    OPEN_SCOPE_INJECTION_PILO = "OpenScopeInjectionPilot"
    SELECT = "Select..."
    SURGERY_X = "SurgeryX"
    T301T = "T301t"
    TASK_TRAINED_NETWORKS_MUL = "TaskTrainedNetworksMultiscope"
    TASK_TRAINED_NETWORKS_NEU = "TaskTrainedNetworksNeuropixel"
    TINY_BLUE_DOT_BEHAVIOR = "TinyBlueDotBehavior"
    VARIABILITY_AIM1 = "VariabilityAim1"
    VARIABILITY_AIM1_PILOT = "VariabilityAim1Pilot"
    VARIABILITY_SPONTANEOUS = "VariabilitySpontaneous"
    VIP_AXONAL_V1_PHASE1 = "VipAxonalV1Phase1"
    VIP_SOMATIC_V1_MESO = "VIPSomaticV1Meso"
    VIP_SOMATIC_V1_PHASE1 = "VipSomaticV1Phase1"
    VIP_SOMATIC_V1_PHASE2 = "VipSomaticV1Phase2"


class LimsRequired(Enum, metaclass=OptionalEnumMeta):
    """Enum class for LimsRequired"""

    NO = "No"
    SELECT = "Select..."
    YES = "Yes"


class Limstaskflow(Enum, metaclass=OptionalEnumMeta):
    """Enum class for Limstaskflow"""

    AIND_EPHYS_PASSIVE_BEHAVI = "AIND Ephys (Passive Behavior)"
    AIND_EPHYS_SURGERY_ONLY = "AIND Ephys (Surgery only)"
    AIND_U19_THALAMUS = "AIND U19 Thalamus"
    BRAIN_LARGE_SCALE_RECORDI = "BRAIN Large Scale Recording"
    BRAIN_MOUSE_BRAIN_CELL_AT = "BRAIN Mouse Brain Cell Atlas trans-synaptic"
    BRAIN_OBSERVATORY_TRANSGE = "Brain Observatory Transgenic Characterization"
    BTV_BRAIN_VIRAL_STRATEGIE = "BTV BRAIN Viral Strategies"
    CITRIC_ACID_PILOT = "Citric Acid Pilot"
    EPHYS_TASK_DEV_DYNAMIC_RO = "Ephys Task Dev Dynamic Routing (NSB Beh)"
    EPHYS_TASK_DEV_DYNAMI_001 = "Ephys Task Dev Dynamic Routing (S/C Beh)"
    EPHYS_TAS_DEV_DYNAMIC_ROU = "Ephys Tas Dev Dynamic Routing DOC Left Eye"
    IVSCC_HVA_RETRO_PATCH_SEQ = "IVSCC HVA Retro PatchSeq"
    IVSC_CM_INJECTION = "IVSCCm injection"
    IVSP_CM_INJECTION = "IVSPCm Injection"
    MGT_ANTEROGRADE_PROJECTIO = "MGT Anterograde Projection Mapping"
    MGT_LAB = "MGT Lab"
    MGT_TISSUE_CYTE = "MGT TissueCyte"
    MSP_DYNAMIC_ROUTING_SURGI = "MSP Dynamic Routing Surgical Development"
    MSP_DYNAMIC_ROUTING_TASK = "MSP Dynamic Routing Task 1 Production"
    MSP_DYNAMIC_ROUTING_ULTRA = (
        "MSP Dynamic Routing Ultra Optotagging Behavior"
    )
    MSP_LEARNING_M_FISH_DEVEL = "MSP Learning & mFISH Development (Dox)"
    MSP_LEARNING_M_FISH_D_001 = "MSP Learning & mFISH Development"
    MSP_LEARNING_M_FISH_FRONT = "MSP Learning & mFISH Frontal Window Dev"
    MSP_LEARNING_M_FISH_VIRUS = "MSP Learning & mFISH Virus Testing"
    MSP_OM_FISH_CO_REGISTRATI = "MSP omFISH Co-Registration Pilot"
    MSP_OM_FISH_CUX2_PILOT = "MSP omFISH Cux2 Pilot"
    MSP_OM_FISH_GAD2_PILOT = "MSP omFISH Gad2 Pilot"
    MSP_OM_FISH_ROB_INJECTION = "MSP omFISH ROB Injection Virus Pilot"
    MSP_OM_FISH_RORB_PILOT = "MSP omFISH Rorb Pilot"
    MSP_OPEN_SCOPE_DENDRITE_C = "MSP OpenScope Dendrite Coupling"
    MSP_OPEN_SCOPE_GLOBAL_001 = (
        "MSP OpenScope Global Local Oddballs (Cohort 2)"
    )
    MSP_OPEN_SCOPE_GLOBAL_LOC = (
        "MSP OpenScope Global Local Oddballs (Cohort 1)"
    )
    MSP_OPEN_SCOPE_ILLUSION = "MSP OpenScope Illusion"
    MSP_TASK_TRAINED_NETWORKS = "MSP Task Trained Networks Neuropixel"
    MSP_TASK_TRAINED_NETW_001 = "MSP Task Trained Networks Multiscope"
    MSP_VARIABILITY_AIM_1 = "MSP Variability Aim 1"
    MSP_VARIABILITY_AIM_1_PIL = "MSP Variability Aim 1 Pilot"
    MSP_VARIABILITY_SPONTANEO = "MSP Variability Spontaneous"
    MSP_VIP_AXONAL_V1 = "MSP VIP Axonal V1"
    MSP_VIP_SOMATIC_V1 = "MSP VIP Somatic V1"
    NA = "N/A"
    OPENSCOPE_VIRUS_VALIDATIO = "Openscope Virus Validation"
    TINY_BLUE_DOT_BEHAVIOR = "Tiny Blue Dot Behavior"
    TRANSGENIC_CHARACTERIZATI = "Transgenic Characterization (Passive)"
    VGT_ENHANCERS_TRANSSYNAPT = "VGT Enhancers Transsynaptic"


class PreviouslyInjected(Enum, metaclass=OptionalEnumMeta):
    """Enum class for PreviouslyInjected"""

    NO = "No"
    SELECT = "Select..."
    YES = "Yes"


class Procedure(Enum, metaclass=OptionalEnumMeta):
    """Enum class for Procedure"""

    CUSTOM = "Custom"
    HPC_CAM = "HP+C CAM"
    HPC_MULTISCOPE = "HP+C Multiscope"
    HPC_NEUROPIXEL_STYLE = "HP+C Neuropixel style"
    HPINJ = "HP+INJ"
    HP_INJECTION_OPTIC_FIBER = "HP+Injection+Optic Fiber Implant"
    HP_ONLY = "HP only"
    HP_TRANSCRANIAL_FOR_ISI = "HP Transcranial (for ISI)"
    INJHPC = "INJ+HP+C"
    INJWHC_NP = "INJ+WHC NP"
    SELECT = "Select..."
    STEREOTAXIC_INJECTION = "Stereotaxic Injection"
    TRAINING_SURGERY = "Training Surgery"
    WHC_NP = "WHC NP"


class ProjectId(Enum, metaclass=OptionalEnumMeta):
    """Enum class for ProjectId"""

    CAMC506 = "CAM-C506"
    CONT503 = "CON-T503"
    CTYT504 = "CTY-T504"
    LOCC500 = "LOC-C500"
    N_1020100910_CTY_MORPHOLO = "102-01-009-10: CTY-Morphology"
    N_1020101010_CTY_SYNAPTIC = "102-01-010-10: CTY-Synaptic Phys"
    N_1020101410_CTY_GEN_TOOL = "102-01-014-10: CTY GenTools Mouse"
    N_1020102120_CTYBRAIN_001 = (
        "102-01-021-20: CTY-BRAIN Mouse Brain Cell Atlas Seg 1"
    )
    N_1020102120_CTYBRAIN_MOU = (
        "102-01-021-20: CTY-BRAIN Mouse Brain Cell Atlas Seg 2 Anatomy"
    )
    N_1020103620_DISSEMINATIO = (
        "102-01-036-20 Dissemination of 3-photon Imaging"
    )
    N_1020104320_OPTICAL_INTE = (
        "102-01-043-20 Optical Interrogation of Venular Function"
    )
    N_1020104410_CTY_GENOMICS = "102-01-044-10 CTY Genomics"
    N_1020104510_IVSCC = "102-01-045-10 IVSCC"
    N_1020104720_3_P_NHP_CORT = "102-01-047-20 3P NHP cortical imaging R34"
    N_1020104810_CTY_BARCODED = "102-01-048-10 CTY Barcoded Connectomics"
    N_1020105520_CTY_EM_MOTOR = "102-01-055-20 CTY EM Motor Cortex"
    N_1020199910_CTY_PROGRAM = "102-01-999-10 CTY Program Admin"
    N_1020201220_BRAIN_VIRAL = "102-02-012-20: BRAIN Viral Strategies"
    N_1020201720_BRAIN_NEUROP = "102-02-017-20 BRAIN Neuropixels Ultra"
    N_1020400620_OTH_MEASURIN = "102-04-006-20: OTH Measuring Consciousness"
    N_1020400710_TARGETED_CNS = "102-04-007-10 Targeted CNS Gene Therapy"
    N_1020401010_CTY_SR_SLC6 = "102-04-010-10 CTY SR: SLC6A1"
    N_1020401410_CTY_PARKINSO = "102-04-014-10 CTY Parkinsons Disease"
    N_1028800410_XPG_CORE_001 = (
        "102-88-004-10 XPG Core Virus Prod, Task: CVS Production"
    )
    N_1028800410_XPG_CORE_VIR = "102-88-004-10 XPG Core Virus Prod, Task: R&D"
    N_1210100410_VIP_REGULATE = (
        "121-01-004-10 VIP Regulated Stabilized Networks"
    )
    N_1210100710_TASK_TRAINED = "121-01-007-10 Task Trained Networks"
    N_1210100810_NEURAL_ENSEM = "121-01-008-10 Neural Ensemble Variability"
    N_1210101010_V1_OM_FISH = "121-01-010-10 V1 omFISH"
    N_1210101110_DYNAMIC_ROUT = "121-01-011-10 Dynamic Routing"
    N_1210101210_LEARNING_M_F = "121-01-012-10 Learning mFISH"
    N_1210101320_MSP_TEMPLETO = "121-01-013-20 MSP Templeton"
    N_1210101620_BRAIN_OPEN_S = "121-01-016-20 BRAIN OpenScope"
    N_1210102320_MSP_TEMPLETO = (
        "121-01-023-20 MSP Templeton - Testing Theories of Consciousness"
    )
    N_1210199910_MINDSCOPE_MS = "121-01-999-10 Mindscope (MSP) Cross Program"
    N_1220100110_NEURAL_DYNAM = (
        "122-01-001-10 Neural Dynamics Scientific Activities"
    )
    N_1220100220_NEURAL_DYNAM = (
        "122-01-002-20 Neural Dynamics U19- Project 1: Period 1"
    )
    N_1220100220_NEURAL_D_001 = (
        "122-01-002-20 Neural Dynamics U19- Project 2: Period 1"
    )
    N_1229999910_NEURAL_DYNAM = "122-99-999-10 Neural Dynamics Admin (Direct)"
    SELECT = "Select..."


class Scabbing(Enum, metaclass=OptionalEnumMeta):
    """Enum class for Scabbing"""

    MILD = "Mild"
    MODERATE = "Moderate"
    NONE = "None"
    SELECT = "Select..."
    SEVERE = "Severe"


class SecondRoundInjIsoLevel(Enum, metaclass=OptionalEnumMeta):
    """Enum class for SecondRoundInjIsoLevel"""

    N_025 = "0.25"
    N_050 = "0.50"
    N_075 = "0.75"
    N_100 = "1.00"
    N_125 = "1.25"
    N_150 = "1.50"
    N_175 = "1.75"
    N_200 = "2.00"
    N_225 = "2.25"
    N_250 = "2.50"
    N_275 = "2.75"
    N_300 = "3.00+"
    SELECT = "Select..."


class SecondRoundIontoIssue(Enum, metaclass=OptionalEnumMeta):
    """Enum class for SecondRoundIontoIssue"""

    NA = "N/A"
    NO = "No"
    SELECT = "Select..."
    YES = "Yes"


class SecondRoundWorkstation(Enum, metaclass=OptionalEnumMeta):
    """Enum class for SecondRoundWorkstation"""

    SELECT = "Select..."
    SWS_1 = "SWS 1"
    SWS_2 = "SWS 2"
    SWS_3 = "SWS 3"
    SWS_4 = "SWS 4"
    SWS_5 = "SWS 5"
    SWS_6 = "SWS 6"
    SWS_7 = "SWS 7"
    SWS_8 = "SWS 8"
    SWS_9 = "SWS 9"


class Secondroundiontonumber(Enum, metaclass=OptionalEnumMeta):
    """Enum class for Secondroundiontonumber"""

    IONTO_1 = "Ionto #1"
    IONTO_10 = "Ionto #10"
    IONTO_2 = "Ionto #2"
    IONTO_3 = "Ionto #3"
    IONTO_4 = "Ionto #4"
    IONTO_5 = "Ionto #5"
    IONTO_6 = "Ionto #6"
    IONTO_7 = "Ionto #7"
    IONTO_8 = "Ionto #8"
    IONTO_9 = "Ionto #9"
    NA = "N/A"
    SELECT = "Select..."


class Secondroundnanojectnumber(Enum, metaclass=OptionalEnumMeta):
    """Enum class for Secondroundnanojectnumber"""

    NA = "N/A"
    NJ1 = "NJ#1"
    NJ2 = "NJ#2"
    NJ3 = "NJ#3"
    NJ4 = "NJ#4"
    NJ5 = "NJ#5"
    NJ6 = "NJ#6"
    NJ7 = "NJ#7"
    NJ8 = "NJ#8"
    SELECT = "Select..."


class Sex(Enum, metaclass=OptionalEnumMeta):
    """Enum class for Sex"""

    FEMALE = "Female"
    MALE = "Male"
    SELECT = "Select..."


class SinusBleed(Enum, metaclass=OptionalEnumMeta):
    """Enum class for SinusBleed"""

    MILD = "Mild"
    MODERATE = "Moderate"
    NA = "N/A"
    NONE = "None"
    SELECT = "Select..."
    SEVERE = "Severe"


class SurgeryStatus(Enum, metaclass=OptionalEnumMeta):
    """Enum class for SurgeryStatus"""

    INJECTION_PENDING = "Injection Pending"
    NEW = "New"
    NO_SURGERY = "No Surgery"
    PHASE_2_PENDING = "Phase 2 Pending"
    PLANNED_ACUTE = "Planned Acute"
    READY_FOR_FEEDBACK = "Ready for Feedback"
    UNPLANNED_ACUTE = "Unplanned Acute"


class TouchUpStatus(Enum, metaclass=OptionalEnumMeta):
    """Enum class for TouchUpStatus"""

    COMPLETE = "Complete"
    PENDING = "Pending"
    SCHEDULED = "Scheduled"
    SELECT = "Select..."
    UNDER_OBSERVATION = "Under Observation"


class WindowClarity(Enum, metaclass=OptionalEnumMeta):
    """Enum class for WindowClarity"""

    ANTERIOR = "Anterior"
    CENTRAL = "Central"
    LATERAL = "Lateral"
    MEDIAL = "Medial"
    NA = "N/A"
    OTHER_IN_COMMENTS = "Other (In comments)"
    POSTERIOR = "Posterior"
    SELECT = "Select..."


class NSBList(BaseModel, extra=Extra.allow):
    """Fields in Sharepoint list"""

    age_at_injection: Optional[str] = Field(
        default=None,
        alias="Age_x0020_at_x0020_Injection",
        title="Age at Injection 1",
    )
    ap2nd_inj: Optional[str] = Field(
        default=None, alias="AP2ndInj", title="Inj 2 A/P"
    )
    author_id: Optional[int] = Field(default=None, alias="AuthorId")
    breg2_lamb: Optional[str] = Field(
        default=None, alias="Breg2Lamb", title="Bregma to Lambda"
    )
    cage: Optional[str] = Field(default=None, alias="Cage", title="Cage")
    color_tag: Optional[str] = Field(
        default=None, alias="_ColorTag", title="Color Tag"
    )
    com_after1st_inj: Optional[After1StInjection] = Field(
        default=None, alias="ComAfter1stInj", title="After 1st Injection"
    )
    com_after2nd_inj: Optional[After2NdInjection] = Field(
        default=None, alias="ComAfter2ndInj", title="After 2nd Injection"
    )
    com_coplanar: Optional[Laceration] = Field(
        default=None, alias="ComCoplanar", title="Laceration"
    )
    com_damage: Optional[Hematoma] = Field(
        default=None, alias="ComDamage", title="Hematoma"
    )
    com_during1st_inj: Optional[During1StInjection] = Field(
        default=None, alias="ComDuring1stInj", title="During 1st Injection"
    )
    com_during2nd_inj: Optional[During2NdInjection] = Field(
        default=None, alias="ComDuring2ndInj", title="During 2nd Injection"
    )
    com_durotomy: Optional[Durotomy] = Field(
        default=None, alias="ComDurotomy", title="Durotomy"
    )
    com_sinusbleed: Optional[SinusBleed] = Field(
        default=None, alias="ComSinusbleed", title="Sinus Bleed"
    )
    com_swelling: Optional[EdemaSwelling] = Field(
        default=None, alias="ComSwelling", title="Edema (Swelling)"
    )
    com_window: Optional[WindowClarity] = Field(
        default=None, alias="ComWindow", title="Window Clarity"
    )
    compliance_asset_id: Optional[str] = Field(
        default=None, alias="ComplianceAssetId", title="Compliance Asset Id"
    )
    contusion: Optional[Contusion] = Field(
        default=None, alias="Contusion", title="Contusion"
    )
    craniotomy_type: Optional[CraniotomyType] = Field(
        default=None, alias="CraniotomyType", title="Craniotomy Type"
    )
    created: Optional[datetime] = Field(
        default=None, alias="Created", title="Created"
    )
    date1st_injection: Optional[datetime] = Field(
        default=None,
        alias="Date1stInjection",
        title="Date 1st Round Injections",
    )
    date2nd_injection: Optional[datetime] = Field(
        default=None,
        alias="Date2ndInjection",
        title="Date 2nd Round Injections",
    )
    date_of_birth: Optional[datetime] = Field(
        default=None, alias="Date_x0020_of_x0020_Birth", title="Date of Birth"
    )
    date_of_surgery: Optional[datetime] = Field(
        default=None,
        alias="Date_x0020_of_x0020_Surgery",
        title="Date of HP Surgery",
    )
    date_range_end: Optional[datetime] = Field(
        default=None, alias="DateRangeEnd", title="DateRangeEnd"
    )
    date_range_start: Optional[datetime] = Field(
        default=None, alias="DateRangeStart", title="DateRangeStart"
    )
    dv2nd_inj: Optional[str] = Field(
        default=None, alias="DV2ndInj", title="Inj 2 D/V"
    )
    editor_id: Optional[int] = Field(default=None, alias="EditorId")
    end_of_week: Optional[datetime] = Field(
        default=None, alias="End_x0020_of_x0020_Week", title="End of Week"
    )
    exudate_severity: Optional[ExudateSeverity] = Field(
        default=None, alias="Exudate_x0020_Severity", title="Exudate Severity"
    )
    eye_affected: Optional[EyeAffected] = Field(
        default=None, alias="Eye_x0020_Affected", title="Eye Affected"
    )
    eye_issue: Optional[EyeIssue] = Field(
        default=None, alias="Eye_x0020_Issue", title="Eye Issue"
    )
    fiber_implant1: Optional[bool] = Field(
        default=None, alias="FiberImplant1", title="Fiber Implant1"
    )
    fiber_implant1_dv: Optional[str] = Field(
        default=None, alias="FiberImplant1DV", title="FiberImplant1DV"
    )
    fiber_implant2: Optional[bool] = Field(
        default=None, alias="FiberImplant2", title="FiberImplant2"
    )
    fiber_implant2_dv: Optional[str] = Field(
        default=None, alias="FiberImplant2DV", title="FiberImplant2DV"
    )
    field30: Optional[str] = Field(
        default=None, alias="field30", title="Inj1Volperpress"
    )
    field50: Optional[str] = Field(
        default=None, alias="field50", title="Inj2Volperpress"
    )
    first_inj_recovery: Optional[float] = Field(
        default=None, alias="FirstInjRecovery", title="1st Inj Recovery"
    )
    first_injection_iso_durat: Optional[str] = Field(
        default=None,
        alias="FirstInjectionIsoDuration",
        title="1st Round Inj Iso Duration",
    )
    first_injection_weight_af: Optional[str] = Field(
        default=None,
        alias="FirstInjectionWeightAfter",
        title="1st Round Inj Weight After",
    )
    first_injection_weight_be: Optional[str] = Field(
        default=None,
        alias="FirstInjectionWeightBefor",
        title="1st Round Inj Weight Before",
    )
    first_round_ionto_issue: Optional[Firstroundiontoissue] = Field(
        default=None, alias="FirstRoundIontoIssue", title="1stRoundIontoIssue"
    )
    headpost_type: Optional[HeadpostType] = Field(
        default=None, alias="HeadpostType", title="Headpost Type"
    )
    hemisphere2nd_inj: Optional[Inj2Hemisphere] = Field(
        default=None, alias="Hemisphere2ndInj", title="Inj 2 Hemisphere"
    )
    hp_a_p: Optional[str] = Field(
        default=None, alias="HP_x0020_A_x002f_P", title="HP A/P"
    )
    hp_diameter: Optional[str] = Field(
        default=None, alias="HP_x0020_Diameter", title="HP Diameter"
    )
    hp_durotomy: Optional[HpDurotomy] = Field(
        default=None, alias="HPDurotomy", title="HP Durotomy"
    )
    hp_inj: Optional[str] = Field(
        default=None, alias="HP_x0020__x0026__x0020_Inj", title="HP & Inj"
    )
    hp_iso_level: Optional[HpIsoLevel] = Field(
        default=None, alias="HPIsoLevel", title="HP Iso Level"
    )
    hp_loc: Optional[HpLocation] = Field(
        default=None, alias="HpLoc", title="HP Location"
    )
    hp_m_l: Optional[str] = Field(
        default=None, alias="HP_x0020_M_x002f_L", title="HP M/L"
    )
    hp_perf: Optional[HpPerferations] = Field(
        default=None, alias="HpPerf", title="HP Perferations"
    )
    hp_prev_inject: Optional[PreviouslyInjected] = Field(
        default=None, alias="HpPrevInject", title="Previously Injected"
    )
    hp_recovery: Optional[float] = Field(
        default=None, alias="HPRecovery", title="HP Recovery"
    )
    hp_requestor: Optional[str] = Field(
        default=None,
        alias="_x0031_HP_x0020_Requestor_x0020_",
        title="HP Requestor Comments",
    )
    hp_requestor_comments_pla: Optional[str] = Field(
        default=None,
        alias="HPRequestorCommentsPlaintext",
        title="Requestor Comments Plain text",
    )
    hp_surgeon_comments: Optional[str] = Field(
        default=None, alias="HPSurgeonComments", title="HP Surgeon Comments"
    )
    hp_work_station: Optional[HpWorkStation] = Field(
        default=None, alias="HpWorkStation", title="HP Work Station"
    )
    iacuc_protocol: Optional[IacucProtocol] = Field(
        default=None,
        alias="IACUC_x0020_Protocol_x0020__x002",
        title="IACUC Protocol",
        description="Create a secondary List for Protocols and link to drugs",
    )
    id: Optional[int] = Field(default=None, alias="ID", title="ID")
    implant_id_coverslip_type: Optional[ImplantIdCoverslipType] = Field(
        default=None,
        alias="ImplantIDCoverslipType",
        title="Implant ID/Coverslip type",
    )
    inj1_alternating_time: Optional[str] = Field(
        default=None, alias="Inj1AlternatingTime", title="Inj1AlternatingTime"
    )
    inj1_angle_v2: Optional[str] = Field(
        default=None, alias="Inj1Angle_v2", title="Inj1Angle_v2"
    )
    inj1_current: Optional[str] = Field(
        default=None, alias="Inj1Current", title="Inj1Current"
    )
    inj1_lenghtof_time: Optional[str] = Field(
        default=None, alias="Inj1LenghtofTime", title="Inj1LenghtofTime"
    )
    inj1_round: Optional[Inj1Round] = Field(
        default=None, alias="Inj1Round", title="Inj1Round"
    )
    inj1_storage_location: Optional[str] = Field(
        default=None, alias="Inj1StorageLocation", title="Inj1StorageLocation"
    )
    inj1_type: Optional[Inj1Type] = Field(
        default=None, alias="Inj1Type", title="Inj1Type"
    )
    inj1_virus_strain_rt: Optional[str] = Field(
        default=None, alias="Inj1VirusStrain_rt", title="Inj1VirusStrain_rt"
    )
    inj1_vol: Optional[str] = Field(
        default=None, alias="Inj1Vol", title="Inj1Vol"
    )
    inj1angle0: Optional[Inj1Angle] = Field(
        default=None, alias="Inj1angle0", title="Inj1angle"
    )
    inj1volperdepth: Optional[str] = Field(
        default=None, alias="inj1volperdepth", title="inj1volperdepth"
    )
    inj2_alternating_time: Optional[str] = Field(
        default=None, alias="Inj2AlternatingTime", title="Inj2AlternatingTime"
    )
    inj2_angle_v2: Optional[str] = Field(
        default=None, alias="Inj2Angle_v2", title="Inj2Angle_v2"
    )
    inj2_current: Optional[str] = Field(
        default=None, alias="Inj2Current", title="Inj2Current"
    )
    inj2_lenghtof_time: Optional[str] = Field(
        default=None, alias="Inj2LenghtofTime", title="Inj2LenghtofTime"
    )
    inj2_round: Optional[Inj2Round] = Field(
        default=None, alias="Inj2Round", title="Inj2Round"
    )
    inj2_storage_location: Optional[str] = Field(
        default=None, alias="Inj2StorageLocation", title="Inj2StorageLocation"
    )
    inj2_type: Optional[Inj2Type] = Field(
        default=None, alias="Inj2Type", title="Inj2Type"
    )
    inj2_virus_strain_rt: Optional[str] = Field(
        default=None, alias="Inj2VirusStrain_rt", title="Inj2VirusStrain_rt"
    )
    inj2_vol: Optional[str] = Field(
        default=None, alias="Inj2Vol", title="Inj2Vol"
    )
    inj2angle0: Optional[Inj2Angle] = Field(
        default=None, alias="Inj2angle0", title="Inj2angle"
    )
    inj2volperdepth: Optional[str] = Field(
        default=None, alias="inj2volperdepth", title="inj2volperdepth"
    )
    ionto_number_hpinj: Optional[Iontonumberhpinj] = Field(
        default=None, alias="IontoNumberHPINJ", title="IontoNumberHPINJ"
    )
    ionto_number_inj1: Optional[Firstroundiontonumber] = Field(
        default=None, alias="IontoNumberInj1", title="1stRoundIontoNumber"
    )
    ionto_number_inj2: Optional[Secondroundiontonumber] = Field(
        default=None, alias="IontoNumberInj2", title="2ndRoundIontoNumber"
    )
    iso_on: Optional[float] = Field(
        default=None, alias="Iso_x0020_On", title="HP Iso Duration"
    )
    issue: Optional[Issue] = Field(default=None, alias="Issue", title="Issue")
    lab_tracks_group: Optional[str] = Field(
        default=None, alias="LabTracks_x0020_Group", title="LabTracks Group"
    )
    lab_tracks_id: Optional[str] = Field(
        default=None,
        alias="LabTracks_x0020_ID",
        title="LabTracks ID",
        description="Lab Tracks Animal ID",
    )
    lab_tracks_requestor: Optional[str] = Field(
        default=None,
        alias="LabTracks_x0020_Requestor",
        title="LabTracks Requestor",
        description="Who requested the animal in LabTracks",
    )
    li_ms_required: Optional[LimsRequired] = Field(
        default=None,
        alias="LIMs_x0020_Required",
        title="LIMs Required",
        description="Is LIMs Entry required",
    )
    light_cycle: Optional[LightCycle] = Field(
        default=None, alias="Light_x0020_Cycle", title="Light Cycle"
    )
    lim_staskflow1: Optional[Limstaskflow] = Field(
        default=None, alias="LIMStaskflow1", title="LIMStaskflow"
    )
    lims_link: Optional[str] = Field(
        default=None, alias="LIMS_x0020_link", title="LIMS link"
    )
    lims_project_code: Optional[LimsProjectCode] = Field(
        default=None,
        alias="Lims_x0020_Project_x0020_Code",
        title="Lims Project Code",
    )
    long1st_round_inj_cmts: Optional[str] = Field(
        default=None, alias="Long1stRoundInjCmts", title="1stRoundInjComments"
    )
    long2nd_rnd_inj_cmts: Optional[str] = Field(
        default=None, alias="Long2ndRndInjCmts", title="2ndRoundInjComments"
    )
    long_requestor_comments: Optional[str] = Field(
        default=None, alias="LongRequestorComments", title="ReqComments"
    )
    long_surgeon_comments: Optional[str] = Field(
        default=None, alias="LongSurgeonComments", title="HPSurgeonComments"
    )
    ml2nd_inj: Optional[str] = Field(
        default=None, alias="ML2ndInj", title="Inj 2 M/L"
    )
    modified: Optional[datetime] = Field(
        default=None, alias="Modified", title="Modified"
    )
    nanoject_number_inj10: Optional[Firstroundnanojectnumber] = Field(
        default=None,
        alias="NanojectNumberInj10",
        title="1stRoundNanojectNumber",
    )
    nanoject_number_inj2: Optional[Secondroundnanojectnumber] = Field(
        default=None,
        alias="NanojectNumberInj2",
        title="2ndRoundNanojectNumber",
    )
    nd_roung_injection_commen: Optional[str] = Field(
        default=None,
        alias="ndRoungInjectionComments",
        title="2ndRoungInjectionComments",
    )
    pedigree_name: Optional[str] = Field(
        default=None, alias="PedigreeName", title="Pedigree Name"
    )
    procedure: Optional[Procedure] = Field(
        default=None,
        alias="Procedure",
        title="Procedure",
        description="Type of Surgical Procedure",
    )
    project_id_te: Optional[ProjectId] = Field(
        default=None,
        alias="Project_x0020_ID_x0020__x0028_te",
        title="Project ID",
    )
    ret_setting0: Optional[Inj1Retsetting] = Field(
        default=None, alias="retSetting0", title="Inj1retSetting"
    )
    ret_setting1: Optional[Inj2Retsetting] = Field(
        default=None, alias="retSetting1", title="Inj2retSetting"
    )
    round1_inj_isolevel: Optional[FirstRoundInjIsoLevel] = Field(
        default=None,
        alias="Round1InjIsolevel",
        title="1st Round Inj Iso Level",
    )
    round2_inj_isolevel: Optional[SecondRoundInjIsoLevel] = Field(
        default=None,
        alias="Round2InjIsolevel",
        title="2nd Round Inj Iso Level",
    )
    scabbing: Optional[Scabbing] = Field(
        default=None, alias="Scabbing", title="Scabbing"
    )
    second_inj_recover: Optional[float] = Field(
        default=None, alias="SecondInjRecover", title="2nd Inj REcovery"
    )
    second_injection_iso_dura: Optional[str] = Field(
        default=None,
        alias="SecondInjectionIsoDuration",
        title="2nd Round Inj Iso Duration",
    )
    second_injection_weight_a: Optional[str] = Field(
        default=None,
        alias="SecondInjectionWeightAfter",
        title="2nd Round Inj Weight After",
    )
    second_injection_weight_b: Optional[str] = Field(
        default=None,
        alias="SecondInjectionWeightBefore",
        title="2nd Round Inj Weight Before",
    )
    second_round_ionto_issue: Optional[SecondRoundIontoIssue] = Field(
        default=None,
        alias="SecondRoundIontoIssue",
        title="2nd Round Ionto Issue",
    )
    sex: Optional[Sex] = Field(default=None, alias="Sex", title="Sex")
    st_round_injection_commen: Optional[str] = Field(
        default=None,
        alias="stRoundInjectionComments",
        title="1stRoundInjectionComments",
    )
    start_of_week: Optional[datetime] = Field(
        default=None, alias="Start_x0020_Of_x0020_Week", title="Start Of Week"
    )
    surgery_status: Optional[SurgeryStatus] = Field(
        default=None, alias="SurgeryStatus", title="Surgery Status"
    )
    title: Optional[str] = Field(default=None, alias="Title", title="Title")
    touch_up_comp: Optional[datetime] = Field(
        default=None,
        alias="Touch_x0020_Up_x0020__x0020_Comp",
        title="Touch Up  Completed",
    )
    touch_up_status: Optional[TouchUpStatus] = Field(
        default=None,
        alias="Touch_x0020_Up_x0020_Status",
        title="Touch Up Status",
    )
    touch_up_weight: Optional[str] = Field(
        default=None,
        alias="Touch_x0020_Up_x0020_Weight_x002",
        title="Touch Up Weight After",
    )
    ui_version_string: Optional[str] = Field(
        default=None, alias="_UIVersionString", title="Version"
    )
    virus_a_p: Optional[str] = Field(
        default=None, alias="Virus_x0020_A_x002f_P", title="Inj 1 A/P"
    )
    virus_d_v: Optional[str] = Field(
        default=None, alias="Virus_x0020_D_x002f_V", title="Inj 1 D/V"
    )
    virus_hemisphere: Optional[Inj1Hemisphere] = Field(
        default=None, alias="Virus_x0020_Hemisphere", title="Inj 1 Hemisphere"
    )
    virus_m_l: Optional[str] = Field(
        default=None, alias="Virus_x0020_M_x002f_L", title="Inj 1 M/L"
    )
    weight_after_surgery: Optional[str] = Field(
        default=None,
        alias="Weight_x0020_after_x0020_Surgery",
        title="HP Weight after",
    )
    weight_before_surger: Optional[str] = Field(
        default=None,
        alias="Weight_x0020_before_x0020_Surger",
        title="HP Weight before",
    )
    work_station1st_injection: Optional[FirstRoundInjWorkstation] = Field(
        default=None,
        alias="WorkStation1stInjection",
        title="1st Round Inj Workstation",
    )
    work_station2nd_injection: Optional[SecondRoundWorkstation] = Field(
        default=None,
        alias="WorkStation2ndInjection",
        title="2nd Round Workstation",
    )
